import logging
from collections import defaultdict
import re
from . import utils

logger = logging.getLogger(__name__)


def get_fix_mod_from_l10n(mslabel, locmod_df):
    l10n = locmod_df.at[mslabel, 'localization']
    logger.debug('Localizations for %s: %s', mslabel, l10n)
    if l10n:
        l10n.pop('non-localized', None)
        top_loc = max(l10n, key=l10n.get)
        logger.debug('Top localization label for %s: %s', mslabel, top_loc)
        return top_loc


def get_fixed_mod_raw(aa, data_dict, choices=None):
    dist_aa = []
    for ms, v in data_dict.items():
        if choices is None or ms in choices:
            dist_aa.append([v[0], v[1]['peptide'].apply(lambda x: x.count(aa)).sum()])
    utils.internal('Counts for %s: %s', aa, dist_aa)
    top_shift = max(dist_aa, key=lambda tup: tup[1])
    return utils.mass_format(top_shift[0])


def determine_fixed_mods_nonzero(reference, locmod_df, data_dict):
    """Determine fixed modifications in case the reference shift is not at zero.
    Needs localization.
    """
    utils.internal('Localizations for %s: %s', reference, locmod_df.at[reference, 'localization'])
    loc = get_fix_mod_from_l10n(reference, locmod_df)
    label = reference
    data_dict = data_dict.copy()
    while loc is None:
        del data_dict[label]
        label = max(data_dict, key=lambda k: data_dict[k][1].shape[0])
        loc = get_fix_mod_from_l10n(label, locmod_df)
        logger.debug('No luck. Trying %s. Got %s', label, loc)
        if not data_dict:
            break
    return loc


def determine_fixed_mods_zero(aastat_result, data_dict, params_dict):
    """Determine fixed modifications in case the reference shift is at zero.
    Does not need localization.
    """
    fix_mod_zero_thresh = params_dict['fix_mod_zero_thresh']
    min_fix_mod_pep_count_factor = params_dict['min_fix_mod_pep_count_factor']

    fix_mod_dict = {}
    reference = utils.mass_format(0)
    aa_rel = aastat_result[reference][2]
    utils.internal('aa_rel:\n%s', aa_rel)
    candidates = aa_rel[aa_rel < fix_mod_zero_thresh].index
    logger.debug('Fixed mod candidates: %s', candidates)
    for i in candidates:
        candidate_label = get_fixed_mod_raw(i, data_dict)
        if candidate_label != reference:
            # number of peptides with `i` at shift `candidate label` must be higher than ...
            count_cand = data_dict[candidate_label][1][params_dict['peptides_column']].str.contains(i).sum()
            # number of peptides with `i` at shift `reference` by a factor of `min_fix_mod_pep_count_factor`
            count_ref = data_dict[reference][1][params_dict['peptides_column']].str.contains(i).sum()
            # peptide count at candidate shift over # of peptides at reference
            est_ratio = count_cand / len(data_dict[reference][1])
            logger.debug('Peptides with %s: ~%d at %s, ~%d at %s. Estimated pct: %f',
                i, count_ref, reference, count_cand, candidate_label, est_ratio)
            if aastat_result[candidate_label][2][i] > fix_mod_zero_thresh and (
                    est_ratio * 100 > fix_mod_zero_thresh * min_fix_mod_pep_count_factor):
                fix_mod_dict[i] = candidate_label
            else:
                logger.debug('Could not find %s anywhere. Can\'t fix.', i)
        else:
            logger.debug('Reference shift is the best for %s.', i)
    return fix_mod_dict


def determine_fixed_mods(aastat_result, aastat_df, locmod_df, data_dict, params_dict):
    reference = aastat_df.loc[aastat_df['is reference']].index[0]
    if reference == utils.mass_format(0):
        logger.info('Reference bin is at zero shift.')
        fix_mod_dict = determine_fixed_mods_zero(aastat_result, data_dict, params_dict)
    else:
        if locmod_df is None:
            logger.warning('No localization data. '
                'Cannot determine fixed modifications when reference mass shift is non-zero.')
            return {}
        logger.info('Reference bin is at %s. Looking for fixed modification to compensate.', reference)
        loc = determine_fixed_mods_nonzero(reference, locmod_df, data_dict)
        if loc:
            aa, shift = utils.parse_l10n_site(loc)
            fix_mod_dict = {aa: shift}
        else:
            logger.info('No localizations. Stopping.')

    return fix_mod_dict


def recommend_isotope_error(aastat_df, locmod_df, params_dict):
    reference = aastat_df.loc[aastat_df['is reference']].index[0]
    ref_peptides = locmod_df.at[reference, '# peptides in bin']
    logger.debug('%d peptides at reference %s', ref_peptides, reference)
    ref_isotopes = []
    label = reference
    while label:
        label = utils.get_isotope_shift(label, locmod_df)
        ref_isotopes.append(label)
    ref_isotopes.pop()

    i = 0
    for i, label in enumerate(ref_isotopes, 1):
        peps = locmod_df.at[label, '# peptides in bin']
        logger.debug('%d peptides at %s.', peps, label)
        if peps * 100 / ref_peptides < params_dict['recommend isotope threshold']:
            return i - 1
    return i


def recalculate_counts(aa, ms, mods_and_counts, data_dict):
    mods_and_counts[aa].pop(ms, None)
    for i, row in data_dict[ms][1].iterrows():
        seq = row['top isoform'].split('.')[1]
        if row['top_terms'] is not None and ms in row['top_terms']:
            if aa == 'N-term' and seq[1] == '[':
                utils.internal('Reducing count of %s for %s (%s)', seq[0], seq, aa)
                if mods_and_counts[seq[0]].get(ms, 0) > 0:
                    mods_and_counts[seq[0]][ms] -= 1
            elif aa == 'C-term' and seq[-1] == ']':
                res = seq.split('[')[0][-1]
                utils.internal('Reducing count of %s for %s (%s)', res, seq, aa)
                if mods_and_counts[res].get(ms, 0) > 0:
                    mods_and_counts[res][ms] -= 1
            elif seq[:2] == aa + '[':
                utils.internal('Reducing count of N-term for %s', seq)
                if mods_and_counts['N-term'].get(ms, 0) > 0:
                    mods_and_counts['N-term'][ms] -= 1
            elif seq[-1] == ']' and seq.split('[')[0][-1] == aa:
                utils.internal('Reducing count of C-term for %s', seq)
                if mods_and_counts['C-term'].get(ms, 0) > 0:
                    mods_and_counts['C-term'][ms] -= 1


def recalculate_with_isotopes(aa, ms, isotope_rec, mods_and_counts, data_dict, locmod_df):
    logger.debug('Recalculating counts for %s @ %s', aa, ms)
    recalculate_counts(aa, ms, mods_and_counts, data_dict)
    i = 0
    while i < isotope_rec:
        label = utils.get_isotope_shift(ms, locmod_df)
        if label:
            logger.debug('Recalculating %s counts for isotope shift %s', aa, label)
            recalculate_counts(aa, label, mods_and_counts, data_dict)
            i += 1
        else:
            break


def same_residue(isoform):
    return ']{' in isoform or re.search(r'\.{.*?}[A-Z]\[', isoform)


def recalculate_varmods(data_dict, mods_and_counts, params_dict):
    # cancel out already configured modifications
    for site, mod in params_dict['var_mod']:
        ms = utils.find_mass_shift(mod, data_dict, params_dict['prec_acc'])
        if ms:
            if mods_and_counts[site].get(ms, 0) > 0:
                logger.debug('Setting all counts for %s @ %s to zero.', ms, site)
                mods_and_counts[site][ms] = 0
    for ms in data_dict:
        shift, df = data_dict[ms]
        for i, row in df.iterrows():
            if row['top_terms'] is not None and ms in row['top_terms']:
                peptide = row[params_dict['peptides_column']]
                if same_residue(row['top isoform']):  # localization and enabled variable modification on the same residue
                    # this should count towards sum of these shifts, not the localized one
                    pos = row['loc_position'][0]
                    mods = utils.get_var_mods(row, params_dict)
                    utils.internal('%s: extracting %d from %s', row['top isoform'], pos, mods)
                    if pos in mods:
                        vm = mods[pos]
                    elif pos == 1:
                        vm = mods[0]
                    elif pos == len(peptide):
                        vm = mods[pos + 1]
                    else:
                        raise KeyError()
                    aa = peptide[pos - 1]
                    if mods_and_counts[aa].get(ms, 0) > 0:
                        utils.internal('Reducing count of %s at %s', aa, ms)
                        mods_and_counts[aa][ms] -= 1
                    if pos == 1 and mods_and_counts['N-term'].get(ms, 0) > 0:
                        mods_and_counts['N-term'][ms] -= 1
                        utils.internal('Reducing count of N-term at %s', ms)
                    if pos == len(peptide) and mods_and_counts['C-term'].get(ms, 0) > 0:
                        utils.internal('Reducing count of C-term at %s', ms)
                        mods_and_counts['C-term'][ms] -= 1
                    sum_ms = utils.find_mass_shift(vm + shift, data_dict, params_dict['prec_acc'])
                    if sum_ms:
                        mods_and_counts[aa][sum_ms] = mods_and_counts[aa].get(sum_ms, 0) + 1
                        utils.internal('Increasing count of %s at %s', aa, sum_ms)
                        if pos == 1:
                            utils.internal('Increasing count of N-term at %s', sum_ms)
                            mods_and_counts['N-term'][sum_ms] = mods_and_counts['N-term'].get(sum_ms, 0) + 1
                        if pos == len(peptide):
                            utils.internal('Increasing count of C-term at %s', sum_ms)
                            mods_and_counts['C-term'][sum_ms] = mods_and_counts['C-term'].get(sum_ms, 0) + 1


def determine_var_mods(aastat_result, aastat_df, locmod_df, data_dict, params_dict, recommended_fix_mods=None):
    if locmod_df is None:
        logger.info('Cannot recommend variable modifications without localization.')
        return {}
    var_mods = []
    recommended = set()
    multiple = params_dict['multiple_mods']
    if multiple:
        logger.info('Recommending multiple modifications on same residue.')
    else:
        logger.info('Recommending one modification per residue.')
    isotope_rec = recommend_isotope_error(aastat_df, locmod_df, params_dict)
    logger.info('Recommended isotope mass error: %d.', isotope_rec)
    if isotope_rec:
        var_mods.append(('isotope error', isotope_rec))
    reference = aastat_df.loc[aastat_df['is reference']].index[0]
    mods_and_counts = defaultdict(dict)  # dict of AA: shift label: count
    for shift in data_dict:
        if shift == reference:
            continue
        l10n = locmod_df.at[shift, 'localization']
        for k, count in l10n.items():
            if k == 'non-localized':
                continue
            aa, locshift = utils.parse_l10n_site(k)
            if locshift == shift:
                mods_and_counts[aa][shift] = count
    logger.debug('Without isotopes, localization counts are:')

    for k, d in mods_and_counts.items():
        logger.debug('%s: %s', k, d)
    if isotope_rec:
        for aa, dcounts in mods_and_counts.items():
            for shift, count in list(dcounts.items()):
                i = 0
                while i < isotope_rec:
                    label = utils.get_isotope_shift(shift, locmod_df)
                    if label:
                        dcounts[shift] = dcounts.get(shift, 0) + mods_and_counts[aa].get(label, 0)
                        # dcounts.pop(label, None)
                        i += 1
                    else:
                        break
        i = 0
        shift = reference
        while i < isotope_rec:
            label = utils.get_isotope_shift(shift, locmod_df)
            if label:
                logger.debug('Removing all counts for isotope shift %s', label)
                for aa, dcounts in mods_and_counts.items():
                    dcounts[label] = 0
                i += 1
            else:
                break
        logger.debug('With isotopes, localization counts are:')
        for k, d in mods_and_counts.items():
            logger.debug('%s: %s', k, d)

    if recommended_fix_mods:
        logger.debug('Subtracting counts for fixed mods.')
        for aa, shift in recommended_fix_mods.items():
            recalculate_with_isotopes(aa, shift, isotope_rec, mods_and_counts, data_dict, locmod_df)

    if params_dict['var_mod']:
        if not multiple:
            logger.info('Multiple variable modifications are disabled, not recommending {} for variable modifications.'.format(
                utils.format_list(set(x[0] for x in params_dict['var_mod']))))
            for aa, shift in params_dict['var_mod']:
                logger.debug('Removing all counts for %s.', aa)
                for sh in mods_and_counts[aa]:
                    mods_and_counts[aa][sh] = 0

        logger.debug('Subtracting counts for variable mods.')
        recalculate_varmods(data_dict, mods_and_counts, params_dict)

    for i in range(params_dict['variable_mods']):
        logger.debug('Choosing variable modification %d. Counts are:', i + 1)
        for k, d in mods_and_counts.items():
            logger.debug('%s: %s', k, d)
            aa_shifts = {aa: max(dcounts, key=dcounts.get) for aa, dcounts in mods_and_counts.items() if dcounts}
        if mods_and_counts:
            aa_counts = {aa: mods_and_counts[aa][shift] for aa, shift in aa_shifts.items()}
            logger.debug('Best localization counts: %s', aa_shifts)
            logger.debug('Values: %s', aa_counts)
            if aa_shifts:
                top_aa = max(aa_shifts, key=aa_counts.get)
                top_shift = aa_shifts[top_aa]
                top_count = aa_counts[top_aa]
                if top_count < params_dict['min_loc_count']:
                    logger.debug('Localization count too small (%d), stopping.', top_count)
                    break
                recommended.add(top_aa)
                var_mods.append((top_aa, top_shift))
                logger.debug('Chose %s @ %s.', top_shift, top_aa)
                recalculate_with_isotopes(top_aa, top_shift, isotope_rec, mods_and_counts, data_dict, locmod_df)
                if not multiple:
                    logger.debug('Removing all counts for %s.', top_aa)
                    for sh in mods_and_counts[top_aa]:
                        mods_and_counts[top_aa][sh] = 0
    return var_mods



import subprocess
import os
import shutil
from . import AA_stat, utils, io
import argparse
import logging
import sys

"""
Created on Sun Jan 26 15:41:40 2020

@author: julia
"""
OS_PARAMS_DEFAULT = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'open_search.params')

logger = logging.getLogger(__name__)

DICT_AA = {
    'add_G_glycine'       : 'G',
    'add_A_alanine'       : 'A',
    'add_S_serine'        : 'S',
    'add_P_proline'       : 'P',
    'add_V_valine'        : 'V',
    'add_T_threonine'     : 'T',
    'add_C_cysteine'      : 'C',
    'add_L_leucine'       : 'L',
    'add_I_isoleucine'    : 'I',
    'add_N_asparagine'    : 'N',
    'add_D_aspartic_acid' : 'D',
    'add_Q_glutamine'     : 'Q',
    'add_K_lysine'        : 'K',
    'add_E_glutamic_acid' : 'E',
    'add_M_methionine'    : 'M',
    'add_H_histidine'     : 'H',
    'add_F_phenylalanine' : 'F',
    'add_R_arginine'      : 'R',
    'add_Y_tyrosine'      : 'Y',
    'add_W_tryptophan'    : 'W',
    'add_Cterm_peptide'   : 'C-term',
    'add_Nterm_peptide'   : 'N-term',
}


def main():
    pars = argparse.ArgumentParser()
    pars.add_argument('--params', help='CFG file with parameters. If there is no file, AA_stat uses default one. '
                      'An example can be found at https://github.com/SimpleNumber/aa_stat', required=False)
    pars.add_argument('--MSFragger', help='Path to MSFragger .jar file. '
                      'If not specified, MSFRAGGER environment variable is used.')
    pars.add_argument('--dir', help='Directory to store the results. Default value is current directory.', default='.')
    pars.add_argument('-v', '--verbosity', type=int, choices=range(4), default=1, help='Output verbosity.')

    input_spectra = pars.add_mutually_exclusive_group(required=True)
    input_spectra.add_argument('--mgf', nargs='+', help='MGF files to search.', default=None)
    input_spectra.add_argument('--mzml', nargs='+', help='mzML files to search.', default=None)

    pars.add_argument('-db', '--fasta', help='FASTA file with decoys for open search. None: with included MSFragger parameters, '
                      'the database is expected to contain decoys. Default decoy prefix is "rev_".'
                      ' If it differs, do not forget to specify it in AA_stat params file.')
    pars.add_argument('--os-params', help='Custom open search parameters.')
    pars.add_argument('-x', '--optimize-fixed-mods',
                      help='Run multiple searches, automatically determine which fixed modifications to apply.',
                      action='store_true', default=False)
    pars.add_argument('-s', '--skip', help='Skip search if pepXML files exist already. If not specified, '
                      'no steps are skipped. If specified without value, first step may be skipped. '
                      'Value is number of steps to skip. Only works with "-x".',
                      nargs='?', default=0, const=1, type=int)
    pars.add_argument('-je', '--java-executable', default='java')
    pars.add_argument('-ja', '--java-args', default='')

    args = pars.parse_args()

    levels = [logging.WARNING, logging.INFO, logging.DEBUG, utils.INTERNAL]
    logging.basicConfig(format='{levelname:>8}: {asctime} {message}',
                        datefmt='[%H:%M:%S]', level=levels[args.verbosity], style='{')

    if not args.MSFragger:
        args.MSFragger = os.environ.get('MSFRAGGER')
    if not args.MSFragger:
        logger.critical('Please specify --MSFragger or set MSFRAGGER environment variable.')
        sys.exit(1)

    logger.info("Starting MSFragger and AA_stat pipeline.")
    spectra = args.mgf or args.mzml
    spectra = [os.path.abspath(i) for i in spectra]
    working_dir = args.dir


    if args.optimize_fixed_mods:
        logger.debug('Skipping up to %d steps.', args.skip)
        step = 1
        fix_mod_dict = {}
        while True:
            logger.info('Starting step %d.', step)
            fig_data, aastat_table, locmod, data_dict, new_fix_mod_dict, var_mod_dict = run_step_os(
                spectra, 'os_step_{}'.format(step), working_dir, args, change_dict=fix_mod_dict, step=step)

            if new_fix_mod_dict:
                for k, v in new_fix_mod_dict.items():
                    fix_mod_dict.setdefault(k, 0.)
                    fix_mod_dict[k] += data_dict[v][0]
                step += 1
            else:
                break
        try:
            if os.path.isfile(os.path.join(working_dir, 'report.html')):
                logger.debug('Removing existing report.html.')
                os.remove(os.path.join(working_dir, 'report.html'))
            os.symlink(os.path.join('os_step_1', 'report.html'), os.path.join(working_dir, 'report.html'))
        except Exception as e:
            logger.debug('Can\'t create symlink to report: %s', e)
        else:
            logger.debug('Symlink created successfully.')
        logger.info('Stopping after %d steps.', step)
    else:
        logger.info('Running one-shot search.')
        folder_name = ''
        run_step_os(spectra, folder_name, working_dir, args)


def get_pepxml(input_file, d=None):
    initial = os.path.splitext(input_file)[0] + '.pepXML'
    if d is None:
        return initial
    sdir, f = os.path.split(initial)
    return os.path.join(d, f)


def run_os(java, jargs, spectra, msfragger, save_dir, parameters):
    command = [java] + jargs + ['-jar', msfragger, parameters, *spectra]
    logger.debug('Running command: %s', ' '.join(command))
    retval = subprocess.call(command)
    logger.debug('Subprocess returned %s', retval)
    if retval:
        logger.critical('MSFragger returned non-zero code %s. Exiting.', retval)
        sys.exit(retval)
    os.makedirs(save_dir, exist_ok=True)
    for s in spectra:
        pepxml = get_pepxml(s)
        if os.path.normpath(os.path.dirname(pepxml)) != os.path.normpath(save_dir):
            logger.debug('Moving %s to %s', pepxml, save_dir)
            shutil.move(pepxml, get_pepxml(s, save_dir))
        else:
            logger.debug('No need to move pepXML file.')


def create_os_params(output, original=None, mass_shifts=None, fastafile=None):
    original = original or OS_PARAMS_DEFAULT
    with open(output, 'w') as new_params, open(original) as default:
        for line in default:
            key = line.split('=')[0].strip()
            if key == 'database_name' and fastafile:
                new_params.write('database_name = {}\n'.format(fastafile))
            elif mass_shifts and DICT_AA.get(key) in mass_shifts:
                aa = DICT_AA[key]
                new_params.write(key + ' = ' + str(mass_shifts[aa]) + '\n')
            else:
                new_params.write(line)


def run_step_os(spectra, folder_name, working_dir, args, change_dict=None, step=None):
    dir = os.path.abspath(os.path.join(working_dir, folder_name))
    os.makedirs(dir, exist_ok=True)
    os_params_path = os.path.abspath(os.path.join(working_dir, folder_name, 'os.params'))
    create_os_params(os_params_path, args.os_params, change_dict, args.fasta)
    pepxml_names = [get_pepxml(s, dir) for s in spectra]
    run = True
    if step is not None:
        if step <= args.skip:
            run = not all(os.path.isfile(f) for f in pepxml_names)
            logger.debug('On step %d, need to run search: %s', step, run)
        else:
            logger.debug('Can\'t skip step %d, running.', step)
    if run:
        run_os(args.java_executable, args.java_args.split(), spectra, args.MSFragger, dir, os_params_path)
    else:
        logger.info('Skipping search.')
    args.pepxml = pepxml_names
    args.csv = None
    args.dir = dir
    params_dict = io.get_params_dict(args)
    return AA_stat.AA_stat(params_dict, args, step=step)



"""
Created on Thu Oct 24 11:44:50 2019

@author: Julia
"""

import pandas as pd
import numpy as np
from collections import defaultdict, Counter
import logging

from pyteomics import mass
try:
    from pyteomics import cmass
except ImportError:
    cmass = mass
import string
from . import utils, io

logger = logging.getLogger(__name__)


def get_theor_spectrum(peptide, acc_frag, ion_types=('b', 'y'), maxcharge=1,
    aa_mass=mass.std_aa_mass, modifications=None, **kwargs):
    """
    Calculates theoretical spectra in two ways: usual one and in integer format (mz / frag_acc).

    Parameters
    ----------
    peptide : list
        Peptide sequence.
    acc_frag : float
        Fragment mass accuracy in Da.
    ion_types : tuple
        Fragment ion types. ('b', 'y')
    maxcharge: int
        Maximum charge of fragment ion.
    aa_mass: dict
        Amino acid masses
    modifications : dict or None
        Dict of modifications applied to peptide (int position -> float mass shift)

    Returns
    -------
    Returns spectrum in two ways (usual, integer). Usual is a dict with key [ion type, charge] and m/z as a value.
    Integer is a dict, where key is ion type and value is a set of integers (m/z / fragment accuracy).
    """
    if not isinstance(peptide, list):
        raise Exception('peptide is not a list: {!r}'.format(peptide))

    peaks = defaultdict(list)
    theor_set = defaultdict(list)
    aa_mass = aa_mass.copy()
    H = mass.nist_mass['H'][0][0]
    nterm_mod = aa_mass.pop('H-', H)
    OH = H + mass.nist_mass['O'][0][0]
    cterm_mod = aa_mass.pop('-OH', OH)
    if modifications is None:
        modifications = {}
    for ind, pep in enumerate(peptide[:-1]):
        for ion_type in ion_types:
            nterminal = ion_type[0] in 'abc'
            for charge in range(1, maxcharge + 1):
                if ind == 0:
                    if nterminal:
                        mz = cmass.fast_mass2(
                            pep, ion_type=ion_type, charge=charge,
                            aa_mass=aa_mass, **kwargs) + (nterm_mod - H + modifications.get(1, 0.)) / charge
                    else:
                        mz = cmass.fast_mass2(''.join(peptide[1:]), ion_type=ion_type, charge=charge,
                                             aa_mass=aa_mass, **kwargs) + (cterm_mod - OH) / charge + sum(
                                             val for key, val in modifications.items() if key > 1) / charge
                else:
                    if nterminal:
                        mz = peaks[ion_type, charge][-1] + (modifications.get(ind + 1, 0.) + aa_mass[pep]) / charge
                    else:
                        mz = peaks[ion_type, charge][-1] - (modifications.get(ind + 1, 0.) + aa_mass[pep]) / charge
                peaks[ion_type, charge].append(mz)
                theor_set[ion_type].append(int(mz / acc_frag))
    theor_set = {k: set(v) for k, v in theor_set.items()}
    # if modifications:
    #     utils.internal('aa_mass: %s', aa_mass)
    #     utils.internal('Theoretical spectrum with modifications: %s, %s, %s', peptide, modifications, peaks)
    return peaks, theor_set


def RNHS_fast(spectrum_idict, theoretical_set, min_matched, ion_types=('b', 'y')):
    """
    Matches experimental and theoretical spectra in int formats.

    Parameters
    ----------

    spectrum_idict : list
        Experimental spectrum in integer format.  Output of preprocess_spectrum.
    theoretical_set: dict
        A dict where key is ion type and value is a set of integers (m/z / fragment accuracy).
        Output of get_theor_spec function.
    min_matched : int
        Minumum peaks to be matched.
    ion_types : tuple
        Fragment ion types. ('b', 'y')

    Returns
    -------

    Number of matched peaks, score.

    """
    matched = []
    isum = 0
    for ion_type in ion_types:
        match = 0
        for ion in theoretical_set[ion_type]:
            if ion in spectrum_idict:
                match += 1
                isum += spectrum_idict[ion]
        matched.append(match)
    matched_approx = sum(matched)
    if matched_approx >= min_matched:
        return matched_approx, np.prod([np.math.factorial(m) for m in matched]) * isum
    else:
        return 0, 0


_preprocessing_cache = {}
def preprocess_spectrum(reader, spec_id, kwargs, acc=0.01):
    """
    Prepares experimental spectrum for matching, converts experimental spectrum to int format.
    Default settings for preprocessing : maximum peaks is 100, dynamic range is 1000.

    Parameters
    ----------
    reader : file reader
        Spectrum file reader
    spec_id : str
        Spectrum id.

    Returns
    -------
    List of experimental mass spectrum in integer format.
    """
    spectrum = _preprocessing_cache.setdefault((reader, spec_id), {})
    if spectrum:
        # logger.debug('Returning cached spectrum %s', spec_id)
        return spectrum
    # logger.debug('Preprocessing new spectrum %s', spec_id)
    original = reader[spec_id]
    maxpeaks = kwargs.get('maxpeaks', 100)
    dynrange = kwargs.get('dynrange', 1000)

    mz_array = original['m/z array']
    int_array = original['intensity array']
    int_array = int_array.astype(np.float32)

    if dynrange:
        i = int_array > int_array.max() / dynrange
        int_array = int_array[i]
        mz_array = mz_array[i]

    if maxpeaks and int_array.size > maxpeaks:
        i = np.argsort(int_array)[-maxpeaks:]
        j = np.argsort(mz_array[i])
        int_array = int_array[i][j]
        mz_array = mz_array[i][j]

    tmp = (mz_array / acc).astype(int)
    for idx, mt in enumerate(tmp):
        i = int_array[idx]
        spectrum[mt] = max(spectrum.get(mt, 0), i)
        spectrum[mt - 1] = max(spectrum.get(mt - 1, 0), i)
        spectrum[mt + 1] = max(spectrum.get(mt + 1, 0), i)
    return spectrum


def peptide_isoforms(peptide, m, sites, prev_aa, next_aa):
    """
    Parameters
    ----------
    peptide : list
        Peptide sequence
    m: modification label to apply
    sites : set
        Amino acids eligible for modification

    Returns
    -------
    set of lists

    """
    isoforms = []
    if ('N-term' in sites or 'Protein N-term' in sites and prev_aa == '-') and len(peptide[0]) == 1 and peptide[0] not in sites:
        isoforms.append((m + peptide[0],) + tuple(peptide[1:]))
    if ('C-term' in sites or 'Protein C-term' in sites and next_aa == '-') and len(peptide[-1]) == 1 and peptide[-1] not in sites:
        isoforms.append(tuple(peptide[:-1]) + (m + peptide[-1],))
    for ind, a in enumerate(peptide):
        if a in sites:
            isoforms.append(tuple(peptide[:ind]) + (m + a,) + tuple(peptide[ind + 1:]))
    return isoforms


def get_candidates_from_aastat(mass_shifts_table, labels, threshold=1.5):
    """
    Get localization candidates from amono acid statistics.

    Parameters
    ----------
    mass_shifts_table : DataFrame
        DF with amino acid statistics for all mass shifts.
    labels : list
        List of amino acids that should be considered.
    threshold : float
        Threshold to be considered as significantly changed.

    Results
    -------
    Series with mass shift as index and list of candidates as value.
    """

    df = mass_shifts_table.loc[:, labels]
    ms, aa = np.where(df > threshold)
    out = {ms: [] for ms in mass_shifts_table.index}
    for i, j in zip(ms, aa):
        out[df.index[i]].append(df.columns[j])
    return pd.Series(out)


def get_full_set_of_candidates(locmod_df):
    """
    Build list of dicts from all_candidates column taking into account the sums of modification.

    Parameters
    ----------
    locmod_df : DataFrame
        DF with candidates for mass shifts.
    Returns
    -------
    Series
    """
    out = defaultdict(list)
    for ind in locmod_df.index:
        out[ind].append({ind: locmod_df.at[ind, 'all candidates']})
        if isinstance(locmod_df.at[ind, 'sum of mass shifts'], list):
            for pair in locmod_df.at[ind, 'sum of mass shifts']:
                tmp_dict = {}
                tmp_dict[pair[0]] = locmod_df.at[pair[0], 'all candidates']
                if len(pair) > 1:
                    tmp_dict[pair[1]] = locmod_df.at[pair[1], 'all candidates']
                out[ind].append(tmp_dict)
    return pd.Series(out)


def localization_of_modification(ms, ms_label, row, loc_candidates, params_dict, spectra_dict, mass_shift_dict):
    """
    Localizes modification for mass shift in a peptide.
    If two peptides isoforms have the same score, modification counts as 'non-localized'.

    Parameters
    ----------
    ms: float
        mass shift
    ms_label : str
        Label for considered mass shift.
    row : DataFrame row
        Data Frame row for filtered PSMs data.
    loc_candidates : list
        List of dicts with candidates for localization. locmod_df['loc candidates']
    params_dict : dict
        Dict with all parameters.
    spectra_dict : dict
        Keys are filenames and values are Pyteomics readers.
    sum_mod : bool
        True if sum of codifications should be considered.

    Returns
    -------
    Counter of localizations, top isoform, score difference
    """
    mass_dict_0 = mass.std_aa_mass.copy()
    mass_dict_0.update(params_dict['fix_mod'])
    peptide = params_dict['peptides_column']
    prev_aa = params_dict['prev_aa_column']
    next_aa = params_dict['next_aa_column']
    charge = row[params_dict['charge_column']]
    modif_labels = string.ascii_lowercase

    mod_dict = utils.get_var_mods(row, params_dict)
    loc_stat_dict = Counter()

    if params_dict['mzml_files']:
        scan = row[params_dict['spectrum_column']].split('.')[1].lstrip('0')
        spectrum_id = 'controllerType=0 controllerNumber=1 scan=' + scan
    else:
        spectrum_id = row[params_dict['spectrum_column']]
    exp_dict = preprocess_spectrum(spectra_dict[row['file']], spectrum_id, {}, acc=params_dict['frag_acc'],)

    top_score, second_score = 0, 0
    top_isoform = None
    top_terms = None
    for terms in loc_candidates:
        scores = []
        mass_dict = mass_dict_0.copy()
        isoform_part = []
        new_isoform_part = []
        i = 0
        isoforms = []
        sequences = []
        for _ms in terms:
            mod_aa = {modif_labels[i] + aa: mass_shift_dict[_ms] + mass_dict[aa] for aa in params_dict['labels']}
            mass_dict.update(mod_aa)
            mass_dict[modif_labels[i]] = mass_shift_dict[_ms]

            if not isoform_part:  # first modification within this shift (or whole shift)
                isoform_part += peptide_isoforms(list(row[peptide]), modif_labels[i], terms[_ms], row[prev_aa], row[next_aa])
                if _ms == ms_label:
                    # this is the whole-shift modification
                    isoforms += isoform_part
                elif len(terms) == 1:
                    # two equal mass shifts form this mass shift. Apply the second half
                    for p in isoform_part:
                        new_isoform_part += peptide_isoforms(p, modif_labels[i], terms[_ms], row[prev_aa], row[next_aa])
            else:
                # second mass shift
                for p in isoform_part:
                    new_isoform_part += peptide_isoforms(p, modif_labels[i], terms[_ms], row[prev_aa], row[next_aa])
            i += 1
        isoforms += new_isoform_part
        sequences = [list(x) for x in isoforms]
        # utils.internal('Generated %d isoforms for terms %s at shift %s', len(sequences), terms.keys(), ms_label)
        for seq in sequences:
            # utils.internal('seq = %s', seq)
            theor_spec = get_theor_spectrum(seq,
                params_dict['frag_acc'], maxcharge=charge, aa_mass=mass_dict, ion_types=params_dict['ion_types'], modifications=mod_dict)
            scores.append(RNHS_fast(exp_dict, theor_spec[1], params_dict['min_spec_matched'], ion_types=params_dict['ion_types'])[1])
        scores = np.array(scores)
        i = np.argsort(scores)[::-1]
        scores = scores[i]
        sequences = np.array(sequences)[i]
        if scores.size:
            if scores[0] > top_score:
                second_score = top_score
                top_score = scores[0]
                top_isoform = sequences[0]
                top_terms = terms
            if scores.size > 1 and scores[1] > second_score:
                second_score = scores[1]

    if top_isoform is None:
        return loc_stat_dict, None, None, None, None

    if any(all(sites <= {'C-term', 'N-term'} for sites in terms.values())
        for terms in loc_candidates):
        # utils.internal('Injecting unmodified spectra for %s', ms)
        unmod_spec = get_theor_spectrum(list(row[peptide]), params_dict['frag_acc'], maxcharge=charge,
            aa_mass=mass_dict_0, ion_types=params_dict['ion_types'], modifications=mod_dict)
        unmod_score = RNHS_fast(exp_dict, unmod_spec[1], params_dict['min_spec_matched'], ion_types=params_dict['ion_types'])[1]
    else:
        unmod_score = 0

    if top_score == second_score or top_score <= unmod_score:
        utils.internal('top score = %f, second score = %f, unmod score = %f', top_score, second_score, unmod_score)
        loc_stat_dict['non-localized'] += 1
        return loc_stat_dict, None, None, None, None

    mass_dict = mass_dict_0.copy()
    # utils.internal('Top isoform is %s for terms %s (shift %s)', top_isoform, top_terms, ms_label)
    i = 0
    for _ms in top_terms:
        mod_aa = {modif_labels[i] + aa: mass_shift_dict[_ms] + mass_dict[aa] for aa in params_dict['labels']}
        mass_dict.update(mod_aa)
        mass_dict[modif_labels[i]] = mass_shift_dict[_ms]
        i += 1
    for ind, a in enumerate(top_isoform):
        if len(a) > 1:
            if ind == 0:
                loc_stat_dict[utils.format_localization_key('N-term', mass_dict[a[0]])] += 1
            elif ind == len(top_isoform) - 1:
                loc_stat_dict[utils.format_localization_key('C-term', mass_dict[a[0]])] += 1
            loc_stat_dict[utils.format_localization_key(a[1], mass_dict[a[0]])] += 1

    scorediff = (top_score - second_score) / top_score
    # utils.internal('Returning: %s %s %s', loc_stat_dict, ''.join(top_isoform), scorediff)
    top_i = ''.join(top_isoform)
    return loc_stat_dict, top_i, top_terms, scorediff, utils.loc_positions(top_isoform)


def localization(df, ms, ms_label, locations_ms, params_dict, spectra_dict, mass_shift_dict):
    """
    Localizes modification or sum of modifications for mass shift and repeat localization if there are redundant candidates.
    If two peptide isoforms have the same max score, modification counts as 'non-localized'.

    Parameters
    ----------
    df : DataFrame
        DF with filtered peptides for considered mass shift.
    ms: float
        mass shift
    ms_label : str
        Considered mass shift label
    locations_ms :
       locmod_df['loc candidates']
    params_dict : dict
        Dict with all paramenters.
    spectra_dict : dict
        Keys are filenames and values file with mass spectra.

    Returns
    -------
    Counter of localizations.
    """
    logger.info('Localizing %s...', ms_label)
    logger.debug('Localizations: %s', locations_ms)
    if len(locations_ms) < 2 and list(locations_ms[0].values())[0] == set():
        df['localization_count'], df['top isoform'], df['top_terms'], df['localization score'], df['loc_position'] = None, None, None, None, None
    else:
        z = list(zip(*df.apply(lambda x: localization_of_modification(
                    ms, ms_label, x, locations_ms, params_dict, spectra_dict, mass_shift_dict), axis=1)))
        names = ['localization_count', 'top isoform', 'top_terms', 'localization score', 'loc_position']
        dt = {'localization score': np.float32}
        for c, v in zip(names, z):
            df[c] = np.array(v, dtype=dt.get(c, np.object_))
    fname = io.table_path(params_dict['output directory'], ms_label)
    peptide = params_dict['peptides_column']

    mod_aa = string.ascii_lowercase
    mod_dicts = {}
    for pair in locations_ms:
        labels_mod = {}
        i = 0
        for m in pair:
            labels_mod[mod_aa[i]] = m
            i += 1
        mod_dicts[tuple(sorted(pair))] = labels_mod
    columns = ['top isoform', 'localization score', params_dict['spectrum_column']]
    df['top isoform'] = df['top isoform'].fillna(df[peptide])
    df.loc[df.top_terms.notna(), 'mod_dict'] = df.loc[df.top_terms.notna(), 'top_terms'].apply(lambda t: mod_dicts[tuple(sorted(t))])
    df['top isoform'] = df.apply(utils.format_isoform, axis=1, args=(params_dict,))
    df[columns].to_csv(fname, index=False, sep='\t')
    result = df['localization_count'].sum() or Counter()
    logger.debug('Localization result for %s: %s', ms_label, result)
    return {ms_label: result}



import jinja2
import logging
import os
import sys
import re
import warnings
import json
import pkg_resources
from datetime import datetime
import math
import operator
import pandas as pd
import lxml.html
from pyteomics import mass
from . import utils, stats

logger = logging.getLogger(__name__)

def format_unimod_repr(record):
    return '<a href="http://www.unimod.org/modifications_view.php?editid1={0[record_id]}">{0[title]}</a>'.format(record)


def matches(row, ms, sites, params_dict):
    ldict = row['localization_count']
    if 'non-localized' in ldict:
        return False
    for loc in ldict:
        site, shift = utils.parse_l10n_site(loc)
        if shift != ms:
            continue
        for possible_site, possible_position in sites:
            if site == possible_site:
                if possible_position[:3] == 'Any':  # Anywhere, Any C-term, Any N-term
                    return True
                if possible_position == 'Protein N-term' and row[params_dict['prev_aa_column']] == '-':
                    return True
                if possible_position == 'Protein C-term' and row[params_dict['next_aa_column']] == '-':
                    return True
    return False


def format_unimod_info(row, df, params_dict):
    out = []
    for record_id in row['unimod accessions']:
        record = utils.UNIMOD[record_id]
        name = format_unimod_repr(record)
        if 'top isoform' in df:
            sites = {(group['site'], group['position']) for group in record['specificity']}
            matching = df.apply(matches, args=(row.name, sites, params_dict), axis=1).sum()
            total = row['# peptides in bin']
            out.append({'label': '{} ({:.0%} match)'.format(name, matching / total),
                'priority': 1 - matching / total, 'type': 'unimod', 'ref': []})
        else:
            out.append({'label': name, 'priority': 1, 'type': 'unimod', 'ref': []})
    return out


def get_label(table, ms, second=False):
    row = table.loc[ms]
    if len(row['raw info']) == 1:
        if len(row['unimod accessions']) == 1:
            record = utils.UNIMOD[next(iter(row['unimod accessions']))]
            return ('+ ' if second else '') + format_unimod_repr(record)
    return ms


def get_artefact_interpretations(row, mass_shift_data_dict, locmod_df, params_dict):
    out = []
    aa_mass = mass.std_aa_mass.copy()
    aa_mass.update(params_dict['fix_mod'])
    enz = params_dict.get('enzyme')
    df = mass_shift_data_dict[row.name][1]
    peps = df[params_dict['peptides_column']]
    match_aa = set()

    for aa, m in aa_mass.items():
        if abs(abs(row['mass shift']) - m) < params_dict['frag_acc']:
            match_aa.add(aa)
    if not match_aa:
        return []

    if enz:
        cut = set(enz['cut']) & match_aa
        nocut = set(enz.get('nocut', []))
    else:
        cut, nocut = None, set()

    explained = False
    if row['mass shift'] < 0:
        # this can be a loss of any terminal amino acid, or ...
        # an artefact of open search, where the peptide is actually unmodified.
        # in the latter case the amino acid should be an enzyme cleavage site
        if cut:
            # possible artefact
            if enz['sense'] == 'C':
                pct = (
                    (peps.str[0].isin(cut) & ~peps.str[1].isin(nocut)) |  # extra amino acid at N-term
                    peps.str[-2].isin(cut)   # extra amino acid at C-term
                    ).sum() / df.shape[0]
            elif enz['sense'] == 'N':
                pct = (
                    peps.str[1].isin(cut) |
                    (peps.str[-1].isin(cut) & ~peps.str[-2].isin(nocut))
                    ).sum() / df.shape[0]
            else:
                logger.critical('Unknown value of sense in specificity: %s', enz)
                sys.exit(1)

            logger.debug('%.1f%% of peptides in %s %s with %s.',
                pct * 100, row.name, ('start', 'end')[enz['sense'] == 'N'], utils.format_list(cut))
            if pct > params_dict['artefact_thresh']:
                out.append('Search artefact: unmodified peptides with extra {} at {}-terminus ({:.0%} match)'.format(
                    utils.format_list(cut), 'CN'[enz['sense'] == 'C'], pct))
                explained = True
            else:
                logger.debug('Not enough peptide support search artefact interpretation.')
        if not explained:
            if 'top isoform' in df:
                lcount = locmod_df.at[row.name, 'localization']
                pct = (
                    lcount.get(utils.format_localization_key('N-term', row.name), 0) +
                    lcount.get(utils.format_localization_key('C-term', row.name), 0)
                ) / df.shape[0]
                logger.debug('%.1f%% of peptides in %s have terminal localization.', pct * 100, row.name)
                if pct > params_dict['artefact_thresh']:
                    out.append('Loss of ' + utils.format_list(match_aa))
                    if not enz:
                        out[-1] += ' or an open search artefact'
    else:
        # this may be a missed cleavage
        if cut:
            keys = [params_dict['prev_aa_column'], params_dict['next_aa_column']]
            pct = df[keys].apply(
                lambda row: bool(cut.intersection(row[keys[0]] + row[keys[1]])), axis=1).sum() / df.shape[0]
            logger.debug('%.1f%% of peptides in %s have %s as neighbor amino acid.',
                pct * 100, row.name, utils.format_list(cut))
            if pct > params_dict['artefact_thresh']:
                out.append('Possible miscleavage (extra {} at terminus)'.format(utils.format_list(cut)))
            else:
                logger.debug('Not enough peptide support search artefact interpretation.')
    return out


def collect_info(row, table, mass_shift_data_dict, locmod_df, params_dict):
    # Each interpretation is a dict with keys: label, priority, type, ref
    options = [{'label': x, 'priority': 0, 'type': 'artefact', 'ref': []} for x in get_artefact_interpretations(
        row, mass_shift_data_dict, locmod_df, params_dict)]
    options.extend(format_unimod_info(row, mass_shift_data_dict[row.name][1], params_dict))
    if row['isotope index']:
        options.append({'label': 'isotope of {}', 'ref': [row['isotope index']],
            'priority': abs(math.log10(table.at[row['isotope index'], '# peptides in bin'] /
                row['# peptides in bin'] / 8)), 'type': 'isotope'})
    if isinstance(row['sum of mass shifts'], list):
        for terms in row['sum of mass shifts']:
            options.append({'label': '{} {}', 'ref': list(terms), 'type': 'sum',
                'priority': 1 - min(table.at[terms[0], '# peptides in bin'],
                    table.at[terms[1], '# peptides in bin']) / table['# peptides in bin'].max()})
    logger.debug('Raw options for row %s: %s', row.name, options)
    return options


def html_info_item(info):
    return '<span class="info_item {0[type]}" data-ref="{0[ref]}">{0[label]}</span>'.format(info)


def format_info(row, table, char_limit):
    s = row['raw info']
    for d in s:
        if d['type'] == 'isotope':
            d['label'] = d['label'].format(get_label(table, d['ref'][0]))
        if d['type'] == 'sum':
            d['label'] = d['label'].format(get_label(table, d['ref'][0]), get_label(table, d['ref'][1], second=True))
    out = []
    total_len = 0
    for info in sorted(s, key=operator.itemgetter('priority')):
        out.append(html_info_item(info))
        cur_len = len(lxml.html.document_fromstring(info['label']).text_content())
        total_len += cur_len
        utils.internal('Label %s assigned length %d (total %d)', info['label'], cur_len, total_len)
        if total_len > char_limit:
            break
    else:
        return ', '.join(out)

    return ', '.join(out[:1]) + '... <span class="expand_info">(<a class="expand_info_link">expand</a>)</span>'


def format_isoform(isoform):
    out = re.sub(r'([A-Z]\[[+-]?[0-9]+\])', r'<span class="loc">\1</span>', isoform)
    out = re.sub(r'([A-Z]?)\{([+-]?[0-9]+)\}', r'<span class="vmod_loc">\1[\2]</span>', out)
    out = re.sub(r'^([A-Z])\.', r'<span class="nterm"><span class="prev_aa">\1</span>.</span>', out)
    out = re.sub(r'\.([A-Z])$', r'<span class="cterm">.<span class="next_aa">\1</span></span>', out)
    return out


def render_html_report(table_, mass_shift_data_dict, locmod_df, params_dict,
    recommended_fmods, recommended_vmods, vmod_combinations, opposite, save_directory, ms_labels, step=None):
    peptide = params_dict['peptides_column']
    path = os.path.join(save_directory, 'report.html')
    if os.path.islink(path):
        logger.debug('Deleting link: %s.', path)
        os.remove(path)

    if table_ is None:
        with open(path, 'w') as f:
            f.write('No mass shifts found.')
        return
    table = table_.copy()
    labels = params_dict['labels']
    table['raw info'] = table.apply(collect_info, axis=1, args=(table, mass_shift_data_dict, locmod_df, params_dict))
    table['Possible interpretations'] = table.apply(format_info, args=(table, params_dict['html_truncate']), axis=1)
    full_info = json.dumps([', '.join(html_info_item(x)
        for x in sorted(y, key=operator.itemgetter('priority'))) for y in table['raw info']])
    artefact_i = json.dumps([i
            for i, (aa, ms) in enumerate(recommended_vmods)
            if aa != 'isotope error' and any(x['type'] == 'artefact' for x in table.at[ms, 'raw info'])
        ])

    with pd.option_context('display.max_colwidth', 250):
        columns = list(table.columns)
        mslabel = '<a id="binh" href="#">mass shift</a>'
        columns[0] = mslabel
        table.columns = columns
        to_hide = list({'is reference', 'sum of mass shifts', 'isotope index', 'unimod accessions',
            'is isotope', 'unimod candidates', 'raw info'}.intersection(columns))
        table_html = table.style.hide_index().hide_columns(to_hide).applymap(
            lambda val: 'background-color: yellow' if val > 1.5 else '', subset=labels
            ).set_precision(3).apply(
            lambda row: ['background-color: #cccccc' if row['is reference'] else '' for cell in row], axis=1).set_table_styles([
                {'selector': 'tr:hover', 'props': [('background-color', 'lightyellow')]},
                {'selector': 'td, th', 'props': [('text-align', 'center')]},
                {'selector': 'td, th', 'props': [('border', '1px solid black')]}]
            ).format({
                mslabel: '<a href="#">{}</a>'.format(utils.MASS_FORMAT).format,
                '# peptides in bin': '<a href="#">{}</a>'.format}).bar(subset='# peptides in bin', color=stats.cc[2]).render(
            uuid="aa_stat_table")

    peptide_tables = []
    for ms in table.index:
        df = mass_shift_data_dict[ms][1]
        if 'localization score' in df and df['localization score'].notna().any():
            df = df.sort_values(['localization score'], ascending=False).loc[:,
                ['top isoform', 'localization score', params_dict['spectrum_column']]]
            df['localization score'] = df['localization score'].astype(float)
        else:
            dfc = df[[peptide, params_dict['spectrum_column'], params_dict['mods_column']]].copy()
            dfc[peptide] = dfc.apply(utils.get_column_with_mods, axis=1, args=(params_dict,))
            dfc[peptide] = (
                df[params_dict['prev_aa_column']].str[0] + '.' + dfc[peptide] + '.' + df[params_dict['next_aa_column']].str[0])
            df = dfc[[peptide, params_dict['spectrum_column']]]
        peptide_tables.append(df.to_html(
            table_id='peptides_' + ms, classes=('peptide_table',), index=False, escape=False, na_rep='',
            formatters={
                'top isoform': format_isoform,
                peptide: format_isoform,
                'localization score': '{:.2f}'.format}))

    varmod_table_styles = [{'selector': 'th.col_heading', 'props': [('display', 'none')]},
            {'selector': 'th.blank', 'props': [('display', 'none')]},
            {'selector': '.data.row0', 'props': [('font-weight', 'bold')]}]

    if params_dict['fix_mod']:
        d = params_dict['fix_mod'].copy()
        d = utils.masses_to_mods(d)
        fixmod = pd.DataFrame.from_dict(d, orient='index', columns=['value']).T.style.set_caption(
            'Configured, fixed').format(utils.MASS_FORMAT).render(uuid="set_fix_mod_table")
    else:
        fixmod = "Set modifications: none."
    if params_dict['var_mod']:
        varmod = pd.DataFrame.from_records(params_dict['var_mod'], columns=['', 'value']).T.style.set_caption(
            'Configured, variable').format(
            lambda x: utils.mass_format(x) if isinstance(x, float) else x).set_table_styles(
            varmod_table_styles).render(uuid="set_var_mod_table")
    else:
        varmod = None
    if recommended_fmods:
        recmod = pd.DataFrame.from_dict(recommended_fmods, orient='index', columns=['value']).T.style.set_caption(
            'Recommended, fixed').render(uuid="rec_fix_mod_table")
    else:
        recmod = "No fixed modifications recommended."

    if recommended_vmods:
        vmod_comb_i = json.dumps(list(vmod_combinations))
        vmod_comb_val = json.dumps(['This modification is a combination of {} and {}.'.format(*v) for v in vmod_combinations.values()])
        opp_mod_i = json.dumps(opposite)
        opp_mod_v = json.dumps(['This modification negates a fixed modification.\n'
            'For closed search, it is equivalent to set {} @ {} as variable.'.format(
                utils.mass_format(-ms_labels[recommended_vmods[i][1]]), recommended_vmods[i][0]) for i in opposite])
        rec_var_mods = pd.DataFrame.from_records(recommended_vmods, columns=['', 'value']).T.style.set_caption(
            'Recommended, variable').format({'isotope error': '{:.0f}'}).set_table_styles(varmod_table_styles).render(uuid="rec_var_mod_table")
    else:
        rec_var_mods = "No variable modifications recommended."
        vmod_comb_i = vmod_comb_val = opp_mod_i = opp_mod_v = '[]'

    reference = table.loc[table['is reference']].index[0]

    if step is None:
        steps = ''
    else:
        if step != 1:
            prev_url = os.path.join(os.path.pardir, 'os_step_{}'.format(step - 1), 'report.html')
            prev_a = r'<a class="prev steplink" href="{}">Previous step</a>'.format(prev_url)
        else:
            prev_a = ''
        if recommended_fmods:
            next_url = os.path.join(os.path.pardir, 'os_step_{}'.format(step + 1), 'report.html')
            next_a = r'<a class="next steplink" href="{}">Next step</a>'.format(next_url)
        else:
            next_a = ''
        steps = prev_a + '\n' + next_a

    version = pkg_resources.get_distribution('AA_stat').version

    write_html(path, table_html=table_html, peptide_tables=peptide_tables, fixmod=fixmod, varmod=varmod,
        reference=reference, recmod=recmod, rec_var_mod=rec_var_mods, steps=steps, version=version, date=datetime.now(),
        vmod_comb_i=vmod_comb_i, vmod_comb_val=vmod_comb_val, opposite_i=opp_mod_i, opposite_v=opp_mod_v,
        full_info=full_info, artefact_i=artefact_i)


def write_html(path, **template_vars):
    with warnings.catch_warnings():
        if not sys.warnoptions:
            warnings.simplefilter('ignore')
        templateloader = jinja2.PackageLoader('AA_stat', '')
        templateenv = jinja2.Environment(loader=templateloader, autoescape=False)
        template_file = 'report.template'
        template = templateenv.get_template(template_file)

    with open(path, 'w') as output:
        output.write(template.render(template_vars))



import matplotlib
matplotlib.use('Agg')
import pylab as plt
from matplotlib.backends.backend_pdf import PdfPages

import ast
import os
import glob
from configparser import ConfigParser
import multiprocessing as mp
from collections import defaultdict
import logging
import operator
import re

import numpy as np
import pandas as pd
from pyteomics import pepxml, mgf, mzml
from . import utils, stats

AA_STAT_PARAMS_DEFAULT = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'default.cfg')
logger = logging.getLogger(__name__)


def preprocess_df(df, filename, params_dict):
    '''
    Preprocesses DataFrame.

    Parameters
    ----------
    df: DataFrame
        Open search result df.
    filename: str
        Path of initial (pepxml or csv) file
    params_dict: dict
        Dict with all input parameters
    Returns
    -------
    DataFrame
    '''
    logger.debug('Preprocessing %s', filename)
    pp = PdfPages(os.path.join(params_dict['output directory'], filename + '.clustering.pdf'))
    window = params_dict['zero_window']
    zero_bin = 0
    shifts = params_dict['mass_shifts_column']
    if not params_dict['decoy_prefix']:
        isdddict = {}
        for prefix in params_dict['decoy_prefix_list']:
            is_decoy = df[params_dict['proteins_column']].apply(
                lambda s: all(x.startswith(prefix) for x in s))
            isd = is_decoy.sum()
            logger.debug('Trying prefix %s for %s... Found %d decoys.', prefix, filename, isd)
            isdddict[prefix] = isd
        prefix = max(isdddict, key=isdddict.get)
        logger.debug('Selected prefix %s for file %s (%d decoys)', prefix, filename, isdddict[prefix])
    else:
        prefix = params_dict['decoy_prefix']

    df['is_decoy'] = df[params_dict['proteins_column']].apply(lambda s: all(x.startswith(prefix) for x in s))

    if not df['is_decoy'].sum():
        logger.error('No decoy IDs found in %s.', filename)
        if not params_dict['decoy_prefix']:
            logger.error('Configured decoy prefixes are: %s. Check you files or config.',
                ', '.join(params_dict['decoy_prefix_list']))
        else:
            logger.error('Configured decoy prefix is: %s. Check your files or config.', prefix)
        return
    ms, filtered = utils.fdr_filter_mass_shift([None, zero_bin, window], df, params_dict)
    n = filtered.shape[0]
    logger.debug('%d filtered peptides near zero.', n)
    df[shifts] = utils.choose_correct_massdiff(
        df[shifts],
        df[params_dict['measured_mass_column']] - df[params_dict['calculated_mass_column']], params_dict)
    if params_dict['calibration'] == 'off':
        logger.info('Mass calibration is disabled. Skipping.')
    elif params_dict['calibration'] != 'simple':
        if n < params_dict['min_peptides_for_mass_calibration']:
            logger.warning('Skipping mass calibration: not enough peptides near zero mass shift.')
        else:
            to_fit, unit = stats.get_fittable_series(filtered, params_dict)
            # save copies of mass shift column, for use in boolean indexing
            shift_copy = df[shifts].copy()
            old_shifts = filtered[shifts].copy()
            if params_dict['clustering']:
                clustering = stats.clusters(filtered, to_fit, unit, filename, params_dict, pp)
                if clustering is None:
                    filtered_clusters = None
                else:
                    filtered_clusters = stats.filter_clusters(clustering, filtered, to_fit, params_dict)
                if not filtered_clusters:
                    logger.info('Clustering was unsuccesful for %s. Calibrating masses in the whole file.', filename)
                elif len(filtered_clusters) == 1:
                    logger.info('One large cluster found in %s. Calibrating masses in the whole file.', filename)
                    filtered_clusters = None
                else:
                    logger.info('Splitting %s into %d pieces.', filename, len(filtered_clusters))
                    f = plt.figure()
                    for i in filtered_clusters:
                        plt.hist(filtered.loc[to_fit.index].loc[clustering.labels_ == i, shifts], label=i, alpha=0.2, bins=25, density=True)
                    plt.xlabel(shifts)
                    plt.title('Before correction')
                    plt.legend()
                    pp.savefig(f)
                    plt.close()
            else:
                filtered_clusters = None

            if not filtered_clusters:
                slices = [None]
                titles = ['Whole file']
                assigned_masks = [slice(None)]
                filtered_clusters = ['<all>']
            else:
                slices, titles = [], []
                for i in filtered_clusters:
                    slices.append(clustering.labels_ == i)
                    titles.append('Cluster {}'.format(i))
                assigned_masks = stats.get_cluster_masks(filtered_clusters, clustering, df, to_fit, params_dict)
            for c, slice_, title, mask in zip(filtered_clusters, slices, titles, assigned_masks):
                # logger.debug('Slice size for cluster %s is: %s', c, slice_.size if slice_ is not None else None)
                to_fit, unit = stats.get_fittable_series(filtered, params_dict, slice_)
                popt = stats._gauss_fit_slice(to_fit, unit, filename, title, params_dict, pp)

                if unit == 'Da':
                    shift_copy.loc[mask] -= popt[1]
                elif unit == 'ppm':
                    shift_copy.loc[mask] -= popt[1] * df[params_dict['calculated_mass_column']] / 1e6
                else:
                    freq_measured = 1e6 / np.sqrt(utils.measured_mz_series(df.loc[mask], params_dict)) - popt[1]
                    mass_corrected = (((1e6 / freq_measured) ** 2) * df.loc[mask, params_dict['charge_column']] -
                        utils.H * df.loc[mask, params_dict['charge_column']])
                    correction = mass_corrected - df.loc[mask, params_dict['measured_mass_column']]
                    logger.debug('Average systematic mass shift for cluster %s: %f', c, -correction.mean())
                    shift_copy.loc[mask] += correction

            # corrected mass shifts are written back here
            df[shifts] = shift_copy
            filtered[shifts] = df.loc[filtered.index, shifts]

            f = plt.figure()
            floc = filtered.loc[old_shifts.abs() < params_dict['zero_window']]
            sc = plt.scatter(floc[shifts], floc[params_dict['rt_column']],
                c=clustering.labels_ if (params_dict['clustering'] and clustering) else 'k')
            if params_dict['clustering'] and clustering:
                plt.legend(*sc.legend_elements(), title='Clusters')
            plt.xlabel(shifts)
            plt.ylabel(params_dict['rt_column'])
            plt.title('After correction')
            pp.savefig(f)
            plt.close()
            if filtered_clusters != ['<all>']:
                f = plt.figure()
                for i in filtered_clusters:
                    plt.hist(floc.loc[clustering.labels_ == i, shifts], label=i, alpha=0.2, bins=25, density=True)
                plt.xlabel(shifts)
                plt.legend()
                pp.savefig(f)
                plt.close()
    pp.close()
    df['file'] = os.path.splitext(os.path.basename(filename))[0]
    df['check_composition'] = df[params_dict['peptides_column']].apply(lambda x: utils.check_composition(x, params_dict['labels']))
    return df.loc[df['check_composition']]


def read_pepxml(fname, params_dict):
    '''
    Reads pepxml file and preprocess it.
    Parameters
    ----------
    fname: str
        Path to pepxml file
    params_dict: dict
        Dict with all input parameters
    Returns
    -------
    DataFrame
    '''
    logger.debug('Reading %s', fname)
    df = pepxml.DataFrame(fname, read_schema=False, columns=operator.itemgetter(
        'peptides_column', 'proteins_column', 'spectrum_column', 'mass_shifts_column', 'charge_column')(params_dict))
    return preprocess_df(df, fname, params_dict)


def read_csv(fname, params_dict):
    """
    Reads csv file.

    Paramenters
    -----------
    fname : str
        Path to file name.
    params_dict : dict
        Dict with paramenters for parsing csv file.
            `csv_delimiter`, `proteins_column`, `proteins_delimiter`
    Returns
    -------
    A DataFrame of csv file.

    """
    # logger.info('Reading %s', fname)
    df = pd.read_csv(fname, sep=params_dict['csv_delimiter'])
    protein = params_dict['proteins_column']
    if (df[protein].str[0] == '[').all() and (df[protein].str[-1] == ']').all():
        df[protein] = df[protein].apply(ast.literal_eval)
    else:
        df[protein] = df[protein].str.split(params_dict['proteins_delimeter'])
    return preprocess_df(df, fname, params_dict)


def read_mgf(file_path):
    return mgf.IndexedMGF(file_path)


def read_mzml(file_path):  # write this
    return mzml.PreIndexedMzML(file_path)


def read_spectra(args):

    readers = {
        'mgf': read_mgf,
        'mzml': read_mzml,
    }
    out_dict = {}
    for ftype, reader in readers.items():
        filenames = getattr(args, ftype)
        if filenames:
            for filename in filenames:
                name = os.path.split(filename)[-1].split('.')[0]  # write it in a proper way
                out_dict[name] = reader(filename)
    return out_dict


def read_input(args, params_dict):
    """
    Reads open search output, assembles all data in one DataFrame.

    """
    logger.info('Reading input files...')
    readers = {
        'pepxml': read_pepxml,
        'csv': read_csv,
    }
    shifts = params_dict['mass_shifts_column']
    nproc = params_dict['processes']
    if nproc == 1:
        dfs = []
        logger.debug('Reading files in one process.')
        for ftype, reader in readers.items():
            filenames = getattr(args, ftype)
            logger.debug('Filenames [%s]: %s', ftype, filenames)
            if filenames:
                for filename in filenames:
                    # dfs.append(reader(filename, params_dict))
                    dfs.append(reader(filename, params_dict))
    else:
        nfiles = 0
        for ftype, reader in readers.items():
            filenames = getattr(args, ftype)
            if filenames:
                nfiles += len(filenames)
        if nproc > 0:
            nproc = min(nproc, nfiles)
        else:
            nproc = min(nfiles, mp.cpu_count())
        logger.debug('Reading files using %s processes.', nproc)
        pool = mp.Pool(nproc)
        results = []
        for ftype, reader in readers.items():
            filenames = getattr(args, ftype)
            logger.debug('Filenames [%s]: %s', ftype, filenames)
            if filenames:
                for filename in filenames:
                    results.append(pool.apply_async(reader, args=(filename, params_dict)))
        dfs = [r.get() for r in results]
        pool.close()
        pool.join()
    if any(x is None for x in dfs):
        logger.critical('There were errors when reading input.')
        return
    logger.info('Starting analysis...')
    logger.debug('%d dfs collected.', len(dfs))
    data = pd.concat(dfs, axis=0)
    data.index = range(len(data))

    data['bin'] = np.digitize(data[shifts], params_dict['bins'])
    return data


def read_config_file(fname):
    params = ConfigParser(delimiters=('=', ':'), comment_prefixes=('#'), inline_comment_prefixes=('#'))

    params.read(AA_STAT_PARAMS_DEFAULT)
    if fname:
        if not os.path.isfile(fname):
            logger.error('Configuration file not found: %s', fname)
        else:
            params.read(fname)
    else:
        logger.info('Using default parameters for AA_stat.')
    return params


def get_parameters(params):
    """
    Reads paramenters from cfg file to one dict.
    Returns dict.
    """
    params_dict = defaultdict()
    # data
    params_dict['decoy_prefix'] = params.get('data', 'decoy prefix')
    params_dict['decoy_prefix_list'] = re.split(r',\s*', params.get('data', 'decoy prefix list'))
    params_dict['FDR'] = params.getfloat('data', 'FDR')
    params_dict['labels'] = params.get('data', 'labels').strip().split()
    params_dict['rule'] = params.get('data', 'cleavage rule')
    # csv input
    params_dict['csv_delimiter'] = params.get('csv input', 'delimiter')
    params_dict['proteins_delimeter'] = params.get('csv input', 'proteins delimiter')
    params_dict['proteins_column'] = params.get('csv input', 'proteins column')
    params_dict['peptides_column'] = params.get('csv input', 'peptides column')
    params_dict['mass_shifts_column'] = params.get('csv input', 'mass shift column')
    params_dict['score_column'] = params.get('csv input', 'score column')
    params_dict['measured_mass_column'] = params.get('csv input', 'measured mass column')
    params_dict['calculated_mass_column'] = params.get('csv input', 'calculated mass column')
    params_dict['rt_column'] = params.get('csv input', 'retention time column')
    params_dict['next_aa_column'] = params.get('csv input', 'next aa column')
    params_dict['prev_aa_column'] = params.get('csv input', 'previous aa column')
    params_dict['spectrum_column'] = params.get('csv input', 'spectrum column')
    params_dict['charge_column'] = params.get('csv input', 'charge column')
    params_dict['mods_column'] = params.get('csv input', 'modifications column')
    params_dict['score_ascending'] = params.getboolean('csv input', 'score ascending')

    # general
    params_dict['bin_width'] = params.getfloat('general', 'width of bin in histogram')
    params_dict['so_range'] = tuple(float(x) for x in params.get('general', 'open search range').split(','))
    params_dict['walking_window'] = params.getfloat('general', 'shifting window')
    params_dict['FDR_correction'] = params.getboolean('general', 'FDR correction')
    params_dict['processes'] = params.getint('general', 'processes')
    params_dict['zero_window'] = params.getfloat('general', 'zero peak window')
    params_dict['prec_acc'] = params.getfloat('general', 'mass shift tolerance')

    params_dict['zero bin tolerance'] = params.getfloat('general', 'zero shift mass tolerance')
    params_dict['zero min intensity'] = params.getfloat('general', 'zero shift minimum intensity')
    params_dict['min_peptides_for_mass_calibration'] = params.getint('general', 'minimum peptides for mass calibration')

    params_dict['specific_mass_shift_flag'] = params.getboolean('general', 'use specific mass shift window')
    params_dict['specific_window'] = [float(x) for x in params.get('general', 'specific mass shift window').split(',')]

    params_dict['figsize'] = tuple(float(x) for x in params.get('general', 'figure size in inches').split(','))
    params_dict['calibration'] = params.get('general', 'mass calibration')
    params_dict['artefact_thresh'] = params.getfloat('general', 'artefact detection threshold')
    params_dict['html_truncate'] = params.getint('general', 'html info truncation length')

    #clustering
    params_dict['clustering'] = params.getboolean('clustering', 'use clustering')
    params_dict['eps_adjust'] = params.getfloat('clustering', 'dbscan eps factor')
    params_dict['min_samples'] = params.getfloat('clustering', 'dbscan min_samples')
    params_dict['clustered_pct_min'] = params.getfloat('clustering', 'total clustered peptide percentage minimum')
    params_dict['cluster_span_min'] = params.getfloat('clustering', 'cluster span percentage minimum')

    # fit
    params_dict['shift_error'] = params.getint('fit', 'shift error')
    params_dict['max_deviation_sigma'] = params.getfloat('fit', 'standard deviation threshold for sigma')
    params_dict['max_deviation_height'] = params.getfloat('fit', 'standard deviation threshold for height')
    params_dict['fit batch'] = params.getint('fit', 'batch')

    # localization
    params_dict['ion_types'] = tuple(params.get('localization', 'ion type').replace(' ', '').split(','))
    params_dict['frag_acc'] = params.getfloat('localization', 'fragment ion mass tolerance')
    params_dict['candidate threshold'] = params.getfloat('localization', 'frequency threshold')
    params_dict['min_spec_matched'] = params.getint('localization', 'minimum matched peaks')
    params_dict['force_term_loc'] = params.getboolean('localization', 'always try terminal localization')
    params_dict['use_all_loc'] = params.getboolean('localization', 'try all localizations')

    # modifications
    params_dict['variable_mods'] = params.getint('modifications', 'recommend variable modifications')
    params_dict['multiple_mods'] = params.getboolean('modifications', 'recommend multiple modifications on residue')
    params_dict['fix_mod_zero_thresh'] = params.getfloat('modifications', 'fixed modification intensity threshold')
    params_dict['min_fix_mod_pep_count_factor'] = params.getfloat('modifications', 'peptide count factor threshold')
    params_dict['recommend isotope threshold'] = params.getfloat('modifications', 'isotope error abundance threshold')
    params_dict['min_loc_count'] = params.getint('modifications', 'minimum localization count')
    return params_dict


def set_additional_params(params_dict):
    if params_dict['specific_mass_shift_flag']:
        logger.info('Custom bin: %s', params_dict['specific_window'])
        params_dict['so_range'] = params_dict['specific_window'][:]

    elif params_dict['so_range'][1] - params_dict['so_range'][0] > params_dict['walking_window']:
        window = params_dict['walking_window'] / params_dict['bin_width']

    else:
        window = (params_dict['so_range'][1] - params_dict['so_range']) / params_dict['bin_width']
    if int(window) % 2 == 0:
        params_dict['window'] = int(window) + 1
    else:
        params_dict['window'] = int(window)  # should be odd
    params_dict['bins'] = np.arange(params_dict['so_range'][0],
        params_dict['so_range'][1] + params_dict['bin_width'], params_dict['bin_width'])


def get_params_dict(args):
    logger.debug('Received args: %s', args)
    fname = args.params
    outdir = args.dir
    params = read_config_file(fname)
    params_dict = get_parameters(params)
    set_additional_params(params_dict)
    params_dict['output directory'] = outdir
    if args.pepxml:
        fmod, vmod = utils.get_fix_var_modifications(args.pepxml[0], params_dict['labels'])
        params_dict['fix_mod'] = fmod
        params_dict['var_mod'] = utils.format_grouped_keys(utils.group_terminal(vmod), params_dict)
        params_dict['enzyme'] = utils.get_specificity(args.pepxml[0])
    return params_dict


_format_globs = {
    'pepxml': ['*.pepXML', '*.pep.xml'],
    'csv': ['*.csv'],
    'mzml': ['*.mzML'],
    'mgf': ['*.mgf'],
}

def resolve_filenames(args):
    for fformat, gs in _format_globs.items():
        value = getattr(args, fformat)
        if value:
            logger.debug('Received %s list: %s', fformat, value)
            out = []
            for val in value:
                if os.path.isdir(val):
                    for g in gs:
                        files = glob.glob(os.path.join(val, g))
                        logger.debug('Found %d files for glob %s', len(files), g)
                        out.extend(files)
                else:
                    out.append(val)
            logger.debug('Final %s list: %s', fformat, out)
            setattr(args, fformat, out)


def table_path(dir, ms):
    return os.path.join(dir, ms + '.csv')


def save_df(ms, df, save_directory, params_dict):
    peptide = params_dict['peptides_column']
    spectrum = params_dict['spectrum_column']
    prev_aa = params_dict['prev_aa_column']
    next_aa = params_dict['next_aa_column']
    table = df[[peptide, spectrum]].copy()
    peptide1 = df.apply(utils.get_column_with_mods, axis=1, args=(params_dict,))
    table[peptide] = df[prev_aa].str[0] + '.' + peptide1 + '.' + df[next_aa].str[0]
    with open(table_path(save_directory, ms), 'w') as out:
        table.to_csv(out, index=False, sep='\t')


def save_peptides(data, save_directory, params_dict):
    for ms_label, (ms, df) in data.items():
        save_df(ms_label, df, save_directory, params_dict)



import os
import operator
import logging
import pandas as pd
import numpy as np
import warnings
from collections import Counter
import re
import pathlib
import itertools as it
from pyteomics import parser, pepxml, mass

logger = logging.getLogger(__name__)

MASS_FORMAT = '{:+.4f}'
UNIMOD = mass.Unimod(pathlib.Path(os.path.join(os.path.abspath(os.path.dirname(__file__)), 'unimod.xml')).as_uri())
INTERNAL = 5
DIFF_C13 = mass.calculate_mass(formula='C[13]') - mass.calculate_mass(formula='C')
H = mass.nist_mass['H+'][0][0]


def internal(*args, **kwargs):
    """Emit log message with level INTERNAL, which is lower than DEBUG."""
    logger.log(INTERNAL, *args, **kwargs)


def mass_format(mass):
    return MASS_FORMAT.format(mass)


def make_0mc_peptides(pep_list, rule):
    """b, y
    In silico cleaves all peptides with a given rule.

    Parameters
    ----------
    pep_list : Iterable
        An iterable of peptides
    rule : str or compiled regex.
        Cleavage rule in pyteomics format.

    Returns
    -------
    Set of fully cleaved peptides.

    """
    out_set = set()
    for i in pep_list:
        out_set.update(parser.cleave(i, rule))
    return out_set


def fdr_filter_mass_shift(mass_shift, data, params_dict):
    shifts = params_dict['mass_shifts_column']
    ms_shift = data.loc[np.abs(data[shifts] - mass_shift[1]) < mass_shift[2], shifts].mean()

    mask = np.abs(data[shifts] - mass_shift[1]) < 3 * mass_shift[2]
    internal('Mass shift %.3f +- 3 * %.3f', mass_shift[1], mass_shift[2])
    data_slice = data.loc[mask].sort_values(by=[params_dict['score_column'], 'spectrum'],
                                ascending=params_dict['score_ascending']).drop_duplicates(subset=params_dict['peptides_column'])
    internal('%d peptide rows selected for filtering', data_slice.shape[0])
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        df = pepxml.filter_df(data_slice, key=params_dict['score_column'],
            fdr=params_dict['FDR'], reverse=not params_dict['score_ascending'], correction=params_dict['FDR_correction'], is_decoy='is_decoy')
    internal('Filtered data for %s: %d rows', mass_shift, df.shape[0])
    return ms_shift, df


def group_specific_filtering(data, mass_shifts, params_dict):
    """
    Selects window around found mass shift and filters using TDA.
    Window is defined as mean +- sigma.

    Parameters
    ----------
    data : DataFrame
        DF with all open search data.
    mass_shifts: numpy array
        Output of utils.fit_peaks function (poptperr matrix). An array of Gauss fitted mass shift
        parameters and their tolerances. [[A, mean, sigma, A_error, mean_error, sigma_error],...]
    params_dict : dict
        Dict with paramenters for parsing csv file.
        `mass_shifts_column`, `FDR`, `FDR_correction`, `peptides_column`

    Returns
    -------
    Dict with mass shifts (in str format) as key and values is a DF with filtered PSMs.
    """
    logger.info('Performing group-wise FDR filtering...')
    out_data = {}
    for ind, ms in enumerate(mass_shifts):
        if ind != len(mass_shifts) - 1:
            diff = abs(ms[1] - mass_shifts[ind + 1][1])
            width_sum = 3 * (ms[2] + mass_shifts[ind + 1][2])
            if diff < width_sum:
                coef = width_sum / diff
                ms[2] /= coef
                mass_shifts[ind + 1][2] /= coef
                logger.debug('Mass shifts %.3f and %.3f are too close, dividing their sigma by %.4f', ms[1], mass_shifts[ind + 1][1], coef)
        shift, df = fdr_filter_mass_shift(ms, data, params_dict)

        if len(df) > 0:
            #  shift = np.mean(df[shifts]) ###!!!!!!!mean of from  fit!!!!
            out_data[mass_format(shift)] = (shift, df)
    logger.info('# of filtered mass shifts = %s', len(out_data))
    return out_data



def check_composition(peptide, aa_labels):
    '''
    Checks composition of peptides.
    Parameters
    ----------
    peptide: str
        Peptide sequence
    aa_labels: list
        list of acceptable aa.
    Returns
    -------
    True if accebtable, False overwise.
    '''
    return set(peptide) < set(aa_labels)


def get_varmod_combinations(recommended_vmods, values, tolerance):
    logger.debug('Received recommended vmods: %s', recommended_vmods)
    counter = Counter(aa for aa, shift in recommended_vmods)
    eligible = {aa for aa, count in counter.items() if count >= 3}
    out = {}
    if eligible:
        for i, (aa, shift) in enumerate(recommended_vmods):
            if aa == 'isotope error' or aa not in eligible:
                continue
            candidates = [(aac, shiftc) for aac, shiftc in recommended_vmods if aac == aa and shiftc != shift]
            for c1, c2 in it.combinations(candidates, 2):
                if abs(values[c1[1]] + values[c2[1]] - values[shift]) <= tolerance:
                    out[i] = (c1[1], c2[1])
    return out


def get_opposite_mods(fmods, rec_fmods, rec_vmods, values, tolerance):
    fmods = masses_to_mods(fmods)
    for aa, mod in rec_fmods.items():
        if aa in fmods:
            fmods[aa] = fmods[aa] + values[mod]
        else:
            fmods[aa] = values[mod]
    logger.debug('Calculating opposites using effective fixed mod dict: %s', fmods)
    vmod_idx = []
    for aaf, fmod in fmods.items():
        for i, (aav, vmod) in enumerate(rec_vmods):
            if aaf == aav and abs(fmod + values[vmod]) < tolerance:
                vmod_idx.append(i)
    return vmod_idx


def find_isotopes(ms, peptides_in_bin, tolerance=0.01):
    """
    Find the isotopes between mass shifts using mass difference of C13 and C12, information of amino acids statistics as well.

    Paramenters
    -----------

    ms : Series
        Series with mass in str format as index and values float mass shift.
    peptides_in_bin : Series
        Series with # of peptides in each mass shift.
    tolerance : float
        Tolerance for isotop matching.

    Returns
    -------
    DataFrame with 'isotop'(boolean) and 'monoisotop_index' columns.
    """
    out = pd.DataFrame({'isotope': False, 'monoisotop_index': None}, index=ms.index)
    np_ms = ms.to_numpy()
    difference_matrix = np.abs(np_ms.reshape(-1, 1) - np_ms.reshape(1, -1) - DIFF_C13)
    isotop, monoisotop = np.where(difference_matrix < tolerance)
    logger.debug('Found %d potential isotopes.', isotop.sum())
    out.iloc[isotop, 0] = True
    out.iloc[isotop, 1] = out.iloc[monoisotop, :].index
    for i, row in out.iterrows():
        if row['isotope']:
            if peptides_in_bin[i] > peptides_in_bin[row['monoisotop_index']]:
                out.at[i, 'isotope'], out.at[i, 'monoisotop_index'] = False, None
    return out


def get_candidates_from_unimod(mass_shift, tolerance, unimod_df):
    """
    Find modifications for `mass_shift` in Unimod.org database with a given `tolerance`.


    Paramenters
    -----------
    mass_shift : float
        Modification mass in Da.
    tolerance : float
        Tolerance for the search in Unimod db.
    unimod_df : DataFrame
        DF with all unimod mo9difications.

    Returns
    -------
    List  of amino acids.

    """
    ind = abs(unimod_df['mono_mass'] - mass_shift) < tolerance
    sites_set = set()
    accessions = set()
    for i, row in unimod_df.loc[ind].iterrows():
        sites_set.update(s['site'] if s['position'][:3] == 'Any' else s['position'] for s in row['specificity'])
        accessions.add(row['record_id'])
    return sites_set, accessions


def find_mod_sum(x, index, sum_matrix, tolerance):
    """
    Finds mass shift that are sum of given mass shift and other mass shift results in already existing mass shift.

    Parameters
    ----------
    x : float
        Mass shift that considered as a component of a modification.
    index : dict
        Map for mass shift indexes and their values.
    sum_matrix : numpy 2D array
        Matrix of sums for all mass shifts.
    tolerance: float
        Matching tolerance in Da.

    Returns
    -------
    List of tuples.
    """
    rows, cols = np.where(np.abs(sum_matrix - x) < tolerance)
    i = rows <= cols
    if rows.size:
        return list(zip(index[rows[i]], index[cols[i]]))
    return None


def find_sums(ms, tolerance=0.005):
    """
    Finds the sums of mass shifts in Series, if it exists.

    Parameters
    ----------
    ms : Series
        Series with mass in str format as index and values float mass shift.
    tolerance : float
        Matching tolerance in Da.

    Returns
    -------
    Series with pairs of mass shift for all mass shifts.

    """
    zero = mass_format(0.0)
    if zero in ms.index:
        col = ms.drop(zero)
    else:
        col = ms
        logger.info('Zero mass shift not found in candidates.')
    values = col.values
    sum_matrix = values.reshape(-1, 1) + values.reshape(1, -1)
    out = col.apply(find_mod_sum, args=(col.index, sum_matrix, tolerance))
    return out


def apply_var_mods(seq, mods):
    parsed = parser.parse(seq)
    out = []
    for i, aa in enumerate(parsed):
        if i in mods:
            out.append('{{{:+.0f}}}'.format(mods[i]) + aa)
        else:
            out.append(aa)
    seqout = ''.join(out)
    internal('%s + %s = %s', seq, mods, seqout)
    return seqout


def get_column_with_mods(row, params_dict):
    peptide = params_dict['peptides_column']
    mods = get_var_mods(row, params_dict)
    return apply_var_mods(row[peptide], mods)


def format_isoform(row, params_dict):
    ms = row['mod_dict']
    seq = row['top isoform']

    pc, nc, mc = operator.itemgetter('prev_aa_column', 'next_aa_column', 'mods_column')(params_dict)
    prev_aa, next_aa = operator.itemgetter(pc, nc)(row)
    mods = get_var_mods(row, params_dict)
    seq = apply_var_mods(seq, mods)

    sequence = re.sub(r'([a-z])([A-Z])', lambda m: '{}[{:+.0f}]'.format(m.group(2), float(ms[m.group(1)])), seq)
    return '{}.{}.{}'.format(prev_aa[0], sequence, next_aa[0])


def get_fix_var_modifications(pepxml_file, labels):
    fout, vout = {}, []
    p = pepxml.PepXML(pepxml_file, use_index=False)
    mod_list = list(p.iterfind('aminoacid_modification'))
    logger.debug('mod_list: %s', mod_list)
    p.reset()
    term_mods = list(p.iterfind('terminal_modification'))
    logger.debug('term_mods: %s', term_mods)
    p.close()
    for m in mod_list:
        if m['aminoacid'] not in labels:
            continue
        if 'peptide_terminus' in m:
            key = '{}-term {}'.format(m['peptide_terminus'].upper(), m['aminoacid'])
        else:
            key = m['aminoacid']
        if m['variable'] == 'N':
            fout[key] = m['mass']
        else:
            vout.append((key, m['massdiff']))
    for m in term_mods:
        if m['variable'] == 'N':
            if m['terminus'] == 'N':
                fout['H-'] = m['mass']
            else:
                fout['-OH'] = m['mass']
        else:
            key = ('Protein ' if m.get('protein_terminus') == 'Y' else '') + m['terminus'] + '-term'
            vout.append((key, m['massdiff']))
    return fout, vout


def get_specificity(pepxml_file):
    with pepxml.PepXML(pepxml_file, use_index=False) as p:
        s = next(p.iterfind('specificity'))
    logger.debug('Extracted enzyme specificity: %s', s)
    return s


def parse_l10n_site(site):
    aa, shift = site.split('_')
    return aa, shift


def mass_to_mod(label, value, aa_mass=mass.std_aa_mass):
    words = label.split()
    if len(words) > 1:
        # terminal mod
        label = words[-1]
    return value - aa_mass.get(label, 0)


def masses_to_mods(d, fix_mod=None):
    aa_mass = mass.std_aa_mass.copy()
    aa_mass['H-'] = 1.007825
    aa_mass['-OH'] = 17.00274
    if fix_mod:
        aa_mass.update(fix_mod)
    d = {k: mass_to_mod(k, v, aa_mass) for k, v in d.items()}
    if 'H-' in d:
        d['N-term'] = d.pop('H-')
    if '-OH' in d:
        d['C-term'] = d.pop('-OH')
    return d


def get_var_mods(row, params_dict):
    # produce a dict for specific PSM: position (int) -> mass shift (float)
    modifications = row[params_dict['mods_column']]
    peptide = params_dict['peptides_column']
    mass_dict_0 = mass.std_aa_mass.copy()
    mass_dict_0['H-'] = 1.007825
    mass_dict_0['-OH'] = 17.00274
    mass_dict_0.update(params_dict['fix_mod'])
    mod_dict = {}
    if modifications:
        internal('Got modifications for peptide %s: %s', row[peptide], modifications)
    for m in modifications:
        mmass, pos = m.split('@')
        mmass = float(mmass)
        pos = int(pos)
        if pos == 0:
            key = 'H-'
        elif pos == len(row[peptide]) + 1:
            key = '-OH'
        else:
            key = row[peptide][pos-1]
        if abs(mmass - mass_dict_0[key]) > params_dict['frag_acc']:
            # utils.internal('%s modified in %s at position %s: %.3f -> %.3f', key, row[peptide], pos, mass_dict_0[key], mmass)
            mod_dict[pos] = mmass - mass_dict_0[key]
    if mod_dict:
        internal('Final mod dict: %s', mod_dict)
    return mod_dict


def format_grouped_keys(items, params_dict):
    out = []
    for k, td in items:
        if k[1:] == '-term':
            t = k[0]
            if isinstance(td, list):
                keys, values = zip(*td)
                diff = max(values) - min(values)
                label_condition = set(keys) >= set(params_dict['labels'])
                if diff < params_dict['prec_acc'] and label_condition:
                    out.append((k, values[0]))  # arbitrary amino acid, they all have the same modification
                    logger.debug('Collapsing %s-terminal mods.', t)
                else:
                    logger.debug('Not collapsing %s-term dict: diff in values is %.3f, set of labels condition is %ssatisfied',
                        t, diff, '' if label_condition else 'not ')
                    for aa, v in td:
                        out.append((k + ' ' + aa, v))
            else:
                out.append((k, td))
        else:
            out.append((k, td))
    logger.debug('Variable mods with grouped keys: %s', out)
    return out


def group_terminal(items):
    grouped = []
    tg = {}
    for k, v in items:
        prefix, protein, term, aa = re.match(r'((Protein)?(?: )?([NC]-term)?)(?: )?([A-Z])?', k).groups()
        if term is None or aa is None:
            grouped.append((k, v))
        else:
            tg.setdefault(prefix, []).append((aa, v))
    grouped.extend(tg.items())
    logger.debug('Variable mods after grouping: %s', grouped)
    return grouped


def format_mod_dict_str(d):
    if d:
        return ', '.join('{} @ {}'.format(v, k) for k, v in d.items())
    return 'none'


def format_mod_dict(d):
    if d:
        return ', '.join('{} @ {}'.format(mass_format(v), k) for k, v in d.items())
    return 'none'


def format_mod_list(items):
    if items:
        return ', '.join('{} @ {}'.format(v, k) for k, v in items)
    return 'none'


def get_isotope_shift(label, locmod_df):
    isotope = locmod_df[locmod_df['isotope index'] == label]
    if not isotope.shape[0]:
        return
    return isotope[isotope['# peptides in bin'] == isotope['# peptides in bin'].max()].index[0]


def format_localization_key(site, ms):
    if not isinstance(ms, str):
        ms = mass_format(ms)
    return site + '_' + ms


def measured_mz_series(df, params_dict):
    return (df[params_dict['measured_mass_column']] + df[params_dict['charge_column']] * H
        ) / df[params_dict['charge_column']]


def calculated_mz_series(df, params_dict):
    return (df[params_dict['calculated_mass_column']] + df[params_dict['charge_column']] * H
        ) / df[params_dict['charge_column']]


def format_list(lst, sep1=', ', sep2=' or '):
    lst = list(lst)
    if not lst:
        return ''
    if len(lst) == 1:
        return lst[0]
    *most, last = lst
    return sep1.join(most) + sep2 + last


def find_mass_shift(value, data_dict, tolerance):
    s = sorted(data_dict, key=lambda x: abs(value - data_dict[x][0]))
    if abs(data_dict[s[0]][0] - value) <= tolerance:
        return s[0]


def loc_positions(top_isoform):
    return [i for i, a in enumerate(top_isoform, 1) if len(a) > 1]


def choose_correct_massdiff(reported, calculated, params_dict):
    maxdiff = np.abs(reported - calculated).max()
    if maxdiff < params_dict['bin_width'] / 2:
        logger.debug('Maximum mass diff is within bounds: %.4f', maxdiff)
        return reported
    elif maxdiff < params_dict['prec_acc']:
        logger.warning('Reported mass shifts have a high calculation error (%.4f).'
        ' Using own calculations', maxdiff)
        return calculated
    else:
        logger.warning('Reported mass shifts differ from calculated values (up to %.4f).'
        ' Using the reported values. Consider reporting this to the developers.', maxdiff)
        return reported



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(
    name                 = 'AA_stat',
    version              = '2.3',
    description          = '''A utility for validation of peptide identification results in proteomics using amino acid counting.''',
    long_description     = (''.join(open('README.MD').readlines())),
    long_description_content_type="text/markdown",
    author               = 'Julia Bubis & Lev Levitsky',
    author_email         = 'julia.bubis@gmail.com',
    install_requires     = ['pyteomics>4.4.1', 'pandas', 'seaborn', 'scipy', 'numpy', 'lxml', 'jinja2', 'scikit-learn'],
    classifiers          = ['Intended Audience :: Science/Research',
                            'Programming Language :: Python :: 3',
                            'Topic :: Scientific/Engineering :: Bio-Informatics',
                            'Topic :: Scientific/Engineering :: Chemistry',
                            'Topic :: Scientific/Engineering :: Physics'],
    license              = 'License :: OSI Approved :: Apache Software License',
    packages             = find_packages(),
    package_data         = {'AA_stat': ['report.template', 'open_search.params', 'default.cfg', 'unimod.xml']},
    entry_points         = {'console_scripts': ['AA_stat=AA_stat.main:main', 'AA_search=AA_stat.aa_search:main']}
    )



import matplotlib
matplotlib.use('Agg')
import pylab as plt

import math
import os
import logging
import warnings
import multiprocessing as mp

import numpy as np
from sklearn import cluster
from scipy.optimize import curve_fit
from scipy.signal import argrelextrema, savgol_filter
import seaborn as sb

from . import utils


logger = logging.getLogger(__name__)
logging.getLogger('matplotlib.font_manager').disabled = True
logging.getLogger('matplotlib.category').disabled = True

cc = ["#FF6600",
      "#FFCC00",
      "#88AA00",
      "#006688",
      "#5FBCD3",
      "#7137C8",
      ]
sb.set_style('white')
colors = sb.color_palette(palette=cc)


def _gauss_fit_slice(to_fit, unit, filename, title, params_dict, mpl_back):
    logger.debug('Fitting zero-shift peptides...')
    f = plt.figure()
    hist_0 = np.histogram(to_fit, bins=int(params_dict['zero_window'] / params_dict['bin_width']))
    hist_y = hist_0[0]
    hist_x = 0.5 * (hist_0[1][:-1] + hist_0[1][1:])
    plt.plot(hist_x, hist_y, 'b+')
    popt, perr = gauss_fitting(max(hist_y), hist_x, hist_y)
    plt.scatter(hist_x, gauss(hist_x, *popt), label='Gaussian fit')
    plt.xlabel('massdiff, ' + unit)
    plt.title(title)
    mpl_back.savefig(f)
    plt.close()
    logger.info('Systematic shift is %.4f %s for file %s [ %s ]', popt[1], unit, filename, title)
    return popt


def clusters(df, to_fit, unit, filename, params_dict, mpl_back):
    if to_fit.shape[0] < 500:
        logger.warning('Not enough data for cluster analysis. Need at least 500 peptides near zero, found %d.', to_fit.shape[0])
        return None
    X = np.empty((to_fit.shape[0], 2))
    X[:, 0] = to_fit
    X[:, 1] = df.loc[to_fit.index, params_dict['rt_column']]
    logger.debug('Clustering a %s array.', X.shape)
    logger.debug('Initial dimensions: %s to %s', X.min(axis=0), X.max(axis=0))
    logger.debug('Converting to square...')
    span_0 = X[:, 0].max() - X[:, 0].min()
    span_1 = X[:, 1].max() - X[:, 1].min()
    ratio = span_1 / span_0
    X[:, 0] *= ratio
    logger.debug('Transformed dimensions: %s to %s', X.min(axis=0), X.max(axis=0))

    eps = span_1 * params_dict['zero_window'] * params_dict['eps_adjust']
    logger.debug('Using eps=%f', eps)
    clustering = cluster.DBSCAN(eps=eps, min_samples=params_dict['min_samples']).fit(X)
    f = plt.figure()
    sc = plt.scatter(to_fit, X[:, 1], c=clustering.labels_)
    plt.legend(*sc.legend_elements(), title='Clusters')
    plt.xlabel(unit)
    plt.ylabel(params_dict['rt_column'])
    mpl_back.savefig(f)
    plt.close()
    f = plt.figure()
    for c in np.unique(clustering.labels_):
        plt.hist(X[clustering.labels_ == c, 1], label=c, alpha=0.5)
    plt.xlabel(params_dict['rt_column'])
    plt.legend()
    mpl_back.savefig(f)
    plt.close()
    return clustering


def cluster_time_span(clustering, label, df, to_fit, params_dict):
    times = df.loc[to_fit.index].loc[clustering.labels_ == label, params_dict['rt_column']]
    return times.min(), times.max()


def span_percentage(span, df, to_fit, params_dict):
    start, end = span
    all_rt = df[params_dict['rt_column']]
    return (end - start) / (all_rt.max() - all_rt.min())


def cluster_time_percentage(clustering, label, df, to_fit, params_dict):
    span = cluster_time_span(clustering, label, df, to_fit, params_dict)
    return span_percentage(span, df, to_fit, params_dict)


def filter_clusters(clustering, df, to_fit, params_dict):
    nclusters = clustering.labels_.max() + 1
    logger.debug('Found %d clusters, %d labels assigned.', nclusters, clustering.labels_.size)
    if not nclusters:
        return []

    out = []
    clustered_peps = 0
    for i in np.unique(clustering.labels_):
        if i == -1:
            continue
        npep = (clustering.labels_ == i).sum()
        if npep < params_dict['min_peptides_for_mass_calibration']:
            logger.debug('Cluster %s is too small for calibration (%d), discarding.', i, npep)
            continue
        span_pct = cluster_time_percentage(clustering, i, df, to_fit, params_dict)
        if span_pct < params_dict['cluster_span_min']:
            logger.debug('Cluster %s spans %.2f%% of the run (too small, thresh = %.2f%%). Discarding.',
                i, span_pct * 100, params_dict['cluster_span_min'] * 100)
            continue
        out.append(i)
        clustered_peps += npep

    logger.debug('Pre-selected clusters: %s', out)
    logger.debug('%.2f%% peptides in clusters, threshold is %.2f%%.',
        clustered_peps / df.shape[0] * 100, params_dict['clustered_pct_min'] * 100)
    if clustered_peps / df.shape[0] < params_dict['clustered_pct_min']:
        logger.debug('Too few peptides in clusters, discarding clusters altogether.')
        return []
    return out


def get_fittable_series(df, params_dict, mask=None):
    window = params_dict['zero_window']
    shifts = params_dict['mass_shifts_column']
    loc = df[shifts].abs() < window
    # logger.debug('loc size for zero shift: %s', loc.size)
    if params_dict['calibration'] == 'gauss':
        to_fit = df.loc[loc, shifts]
        unit = 'Da'
    elif params_dict['calibration'] == 'gauss_relative':
        to_fit = df.loc[loc, shifts] * 1e6 / df.loc[loc, params_dict['calculated_mass_column']]
        unit = 'ppm'
    elif params_dict['calibration'] == 'gauss_frequency':
        freq_measured = 1e6 / np.sqrt(utils.measured_mz_series(df, params_dict))
        freq_calculated = 1e6 / np.sqrt(utils.calculated_mz_series(df, params_dict))
        to_fit = (freq_measured - freq_calculated).loc[loc]
        unit = 'freq. units'
    if mask is not None:
        to_fit = to_fit.loc[mask]
    logger.debug('Returning a %s fittable series for a %s dataframe with a %s mask.', to_fit.shape, df.shape,
        mask.shape if mask is not None else None)
    return to_fit, unit


def get_cluster_masks(filtered_clusters, clustering, df, to_fit, params_dict):
    all_rt = df[params_dict['rt_column']]
    time_spans = {i: cluster_time_span(clustering, i, df, to_fit, params_dict) for i in filtered_clusters}
    sorted_clusters = sorted(filtered_clusters, key=time_spans.get)  # sorts by span start
    i = 0
    prev = all_rt.min()
    masks = {}
    while i < len(sorted_clusters):
        cur_end = time_spans[sorted_clusters[i]][1]
        if i == len(sorted_clusters) - 1:
            next_point = all_rt.max() + 1
        else:
            next_start = time_spans[sorted_clusters[i + 1]][0]
            next_point = (cur_end + next_start) / 2
        logger.debug('Time span %.1f - %.1f assigned to cluster %s', prev, next_point, sorted_clusters[i])
        masks[sorted_clusters[i]] = (all_rt >= prev) & (all_rt < next_point)
        i += 1
        prev = next_point

    assigned_masks = [masks[c] for c in filtered_clusters]
    return assigned_masks


def smooth(y, window_size=15, power=5):
    """
    Smoothes function.
    Paramenters
    -----------
    y : array-like
        function to smooth.
    window_size : int
        Smothing window.
    power : int
        Power of smothing function.

    Returns
    -------
    Smoothed function

    """
    y_smooth = savgol_filter(y, window_size, power)
    return y_smooth


def gauss(x, a, x0, sigma):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        return a / sigma / np.sqrt(2 * np.pi) * np.exp(-(x - x0) * (x - x0) / (2 * sigma ** 2))


def gauss_fitting(center_y, x, y):
    """
    Fits with Gauss function
    `center_y` - starting point for `a` parameter of gauss
    `x` numpy array of mass shifts
    `y` numpy array of number of psms in this mass shifts

    """
    mean = (x * y).sum() / y.sum()
    sigma = np.sqrt((y * (x - mean) ** 2).sum() / y.sum())
    a = center_y * sigma * np.sqrt(2 * np.pi)
    try:
        popt, pcov = curve_fit(gauss, x, y, p0=(a, mean, sigma))
        perr = np.sqrt(np.diag(pcov))
        return popt, perr
    except (RuntimeError, TypeError):
        return None, None


def fit_worker(args):
    return fit_batch_worker(*args)


def fit_batch_worker(out_path, batch_size, xs, ys, half_window, height_error, sigma_error):
    shape = int(math.ceil(np.sqrt(batch_size)))
    figsize = (shape * 3, shape * 4)
    plt.figure(figsize=figsize)
    plt.tight_layout()
    logger.debug('Created a figure with size %s', figsize)
    poptpvar = []
    for i in range(batch_size):
        center = i * (2 * half_window + 1) + half_window
        x = xs[center - half_window: center + half_window + 1]
        y = ys[center - half_window: center + half_window + 1]
        popt, perr = gauss_fitting(ys[center], x, y)
        plt.subplot(shape, shape, i + 1)
        if popt is None:
            label = 'NO FIT'
        else:
            if (x[0] <= popt[1] and popt[1] <= x[-1] and perr[0] / popt[0] < height_error
                    and perr[2] / popt[2] < sigma_error):
                label = 'PASSED'
                poptpvar.append(np.concatenate([popt, perr]))
                plt.vlines(popt[1] - 3 * popt[2], 0, ys[center], label='3sigma interval')
                plt.vlines(popt[1] + 3 * popt[2], 0, ys[center])
            else:
                label = 'FAILED'
        plt.plot(x, y, 'b+:', label=label)
        if label != 'NO FIT':
            plt.scatter(x, gauss(x, *popt), label='Gaussian fit\n $\\sigma$ = {:.4f}'.format(popt[2]))

        plt.legend()
        plt.title("{0:.3f}".format(xs[center]))
        plt.grid(True)

    logger.debug('Fit done. Saving %s...', out_path)
    plt.savefig(out_path)
    plt.close()
    return poptpvar


def fit_peaks(data, args, params_dict):
    """
    Finds Gauss-like peaks in mass shift histogram.

    Parameters
    ----------
    data : DataFRame
        A DF with all (non-filtered) results of open search.
    args: argsparse
    params_dict : dict
        Parameters dict.
    """
    logger.info('Performing Gaussian fit...')
    fit_batch = params_dict['fit batch']
    half_window = int(params_dict['window'] / 2) + 1
    hist = np.histogram(data[data['is_decoy'] == False][params_dict['mass_shifts_column']], bins=params_dict['bins'])
    hist_y = smooth(hist[0], window_size=params_dict['window'], power=5)
    hist_x = 0.5 * (hist[1][:-1] + hist[1][1:])
    loc_max_candidates_ind = argrelextrema(hist_y, np.greater_equal)[0]
    # smoothing and finding local maxima
    min_height = 2 * np.median(hist[0][hist[0] > 1])
    # minimum bin height expected to be peak approximate noise level as median of all non-negative
    loc_max_candidates_ind = loc_max_candidates_ind[hist_y[loc_max_candidates_ind] >= min_height]
    if not loc_max_candidates_ind.size:
        logger.info('No peaks found for fit.')
        return hist, np.array([])
    height_error = params_dict['max_deviation_height']
    sigma_error = params_dict['max_deviation_sigma']
    logger.debug('Candidates for fit: %s', len(loc_max_candidates_ind))
    nproc = int(math.ceil(len(loc_max_candidates_ind) / fit_batch))
    maxproc = params_dict['processes']
    if maxproc > 0:
        nproc = min(nproc, maxproc)
    if nproc > 1:
        arguments = []
        logger.debug('Splitting the fit into %s batches...', nproc)
        n = min(nproc, mp.cpu_count())
        logger.debug('Creating a pool of %s processes.', n)
        pool = mp.Pool(n)
        for proc in range(nproc):
            xlist = [hist_x[center - half_window: center + half_window + 1]
                for center in loc_max_candidates_ind[proc * fit_batch: (proc + 1) * fit_batch]]
            xs = np.concatenate(xlist)
            ylist = [hist[0][center - half_window: center + half_window + 1]
                for center in loc_max_candidates_ind[proc * fit_batch: (proc + 1) * fit_batch]]
            ys = np.concatenate(ylist)
            out = os.path.join(args.dir, 'gauss_fit_{}.pdf'.format(proc + 1))
            arguments.append((out, len(xlist), xs, ys, half_window, height_error, sigma_error))
        res = pool.map_async(fit_worker, arguments)
        poptpvar_list = res.get()
        # logger.debug(poptpvar_list)
        pool.close()
        pool.join()
        logger.debug('Workers done.')
        poptpvar = [p for r in poptpvar_list for p in r]
    else:
        xs = np.concatenate([hist_x[center - half_window: center + half_window + 1]
                for center in loc_max_candidates_ind])
        ys = np.concatenate([hist[0][center - half_window: center + half_window + 1]
                for center in loc_max_candidates_ind])
        poptpvar = fit_batch_worker(os.path.join(args.dir, 'gauss_fit.pdf'),
            len(loc_max_candidates_ind), xs, ys, half_window, height_error, sigma_error)

    logger.debug('Returning from fit_peaks. Array size is %d.', len(poptpvar))
    return hist, np.array(poptpvar)


_Mkstyle = matplotlib.markers.MarkerStyle
_marker_styles = [_Mkstyle('o', fillstyle='full'), (_Mkstyle('o', fillstyle='left'), _Mkstyle('o', fillstyle='right')),
    (_Mkstyle('o', fillstyle='top'), _Mkstyle('o', fillstyle='bottom')), (_Mkstyle(8), _Mkstyle(9)),
    (_Mkstyle('v'), _Mkstyle('^')), (_Mkstyle('|'), _Mkstyle('_')), (_Mkstyle('+'), _Mkstyle('x'))]


def _generate_pair_markers():
    '''Produce style & color pairs for localization markers (except the main one).'''
    for i in [3, 4, 5, 0, 1, 2]:
        for ms in _marker_styles[1:]:
            yield colors[i], ms


def _get_max(arr):
    values = [x for x in arr if x is not None]
    if values:
        return max(values)
    return 0


def plot_figure(ms_label, ms_counts, left, right, params_dict, save_directory, localizations=None, sumof=None):
    """
    Plots amino acid spatistics.

    Parameters
    ----------
    ms_label : str
        Mass shift in string format.
    ms_counts : int
        Number of peptides in a mass shift.
    left : list
        Amino acid statistics data [[values], [errors]]
    right : list
        Amino acid frequences in peptides
     params_dict : dict
        Parameters dict.
    save_directory: str
        Saving directory.
    localizations : Counter
         Localization counter using  ms/ms level.
    sumof : List
        List of str tuples for constituent mass shifts.
    """
    b = 0.1  # shift in bar plots
    width = 0.2  # for bar plots
    labels = params_dict['labels']
    labeltext = ms_label + ' Da mass shift,\n' + str(ms_counts) + ' peptides'
    x = np.arange(len(labels))
    distributions = left[0]
    errors = left[1]

    fig, ax_left = plt.subplots()
    fig.set_size_inches(params_dict['figsize'])

    ax_left.bar(x - b, distributions.loc[labels],
            yerr=errors.loc[labels], width=width, color=colors[2], linewidth=0)

    ax_left.set_ylabel('Relative AA abundance', color=colors[2])
    ax_left.set_xticks(x)
    ax_left.set_xticklabels(labels)
    ax_left.hlines(1, -1, x[-1] + 1, linestyles='dashed', color=colors[2])
    ax_right = ax_left.twinx()

    ax_right.bar(x + b, right, width=width, linewidth=0, color=colors[0])

    ax_right.set_ylim(0, 125)
    ax_right.set_yticks(np.arange(0, 120, 20))
    ax_right.set_ylabel('Peptides with AA, %', color=colors[0])

    ax_left.spines['left'].set_color(colors[2])
    ax_right.spines['left'].set_color(colors[2])

    ax_left.spines['right'].set_color(colors[0])
    ax_right.spines['right'].set_color(colors[0])
    ax_left.tick_params('y', colors=colors[2])
    ax_right.tick_params('y', colors=colors[0])

    pright = matplotlib.lines.Line2D([], [], marker=None, label=labeltext, alpha=0)

    ax_left.set_xlim(-1, x[-1] + 1)
    ax_left.set_ylim(0, distributions.loc[labels].max() * 1.4)

    logger.debug('Localizations for %s figure: %s', ms_label, localizations)
    if localizations:
        ax3 = ax_left.twinx()
        ax3.spines['right'].set_position(('axes', 1.1))
        ax3.set_frame_on(True)
        ax3.patch.set_visible(False)
        ax3.set_ylabel('Localization count', color=colors[3])
        for sp in ax3.spines.values():
            sp.set_visible(False)
        ax3.spines['right'].set_visible(True)
        ax3.spines['right'].set_color(colors[3])
        ax3.tick_params('y', colors=colors[3])
        # plot simple modifications (not sum) with the first style,
        # then parts of sum
        values = [localizations.get(key + '_' + ms_label) for key in labels]
        maxcount = _get_max(values)
        label_prefix = 'Location of '
        ax3.scatter(x, values, marker=_marker_styles[0], color=colors[3], label=label_prefix + ms_label)
        if isinstance(sumof, list):
            for pair, (color, style) in zip(sumof, _generate_pair_markers()):
                values_1 = [localizations.get(key + '_' + pair[0]) for key in labels]
                maxcount = max(maxcount, _get_max(values_1))
                ax3.scatter(x, values_1, marker=style[0], color=color, label=label_prefix + pair[0])
                if pair[0] != pair[1]:
                    values_2 = [localizations.get(key + '_' + pair[1]) for key in labels]
                    if values_2:
                        maxcount = max(maxcount, _get_max(values_2))
                        ax3.scatter(x, values_2, marker=style[1], color=color, label=label_prefix + pair[1])
        terms = {key for key in localizations if key[1:6] == '-term'}
        # logger.debug('Found terminal localizations: %s', terms)
        for t in terms:
            label = '{} at {}: {}'.format(*reversed(t.split('_')), localizations[t])
            p = ax3.plot([], [], label=label)[0]
            p.set_visible(False)
        pright.set_label(pright.get_label() + '\nNot localized: {}'.format(localizations.get('non-localized', 0)))
        if maxcount:
            ax3.legend(loc='upper left', ncol=2)
        ax3.set_ylim(0, 1.4 * max(maxcount, 1))

    ax_right.legend(handles=[pright], loc='upper right', edgecolor='dimgrey', fancybox=True, handlelength=0)
    fig.tight_layout()
    fig.savefig(os.path.join(save_directory, ms_label + '.png'), dpi=500)
    fig.savefig(os.path.join(save_directory, ms_label + '.svg'))
    plt.close()


def summarizing_hist(table, save_directory):
    width = 0.8
    fig, ax = plt.subplots(figsize=(len(table), 5))
    ax.bar(range(len(table)), table.sort_values('mass shift')['# peptides in bin'],
        color=colors[2], align='center', width=width)
    ax.set_title('Peptides in mass shifts', fontsize=12)
    ax.set_xlabel('Mass shift', fontsize=10)
    ax.set_ylabel('Number of peptides')
    ax.set_xlim((-1, len(table)))
    ax.set_xticks(range(len(table)))
    ax.set_xticklabels(table.sort_values('mass shift')['mass shift'].apply('{:.2f}'.format))

    total = table['# peptides in bin'].sum()
    vdist = table['# peptides in bin'].max() * 0.01
    max_height = 0
    for i, patch in enumerate(ax.patches):
        current_height = patch.get_height()
        if current_height > max_height:
            max_height = current_height
        ax.text(patch.get_x() + width / 2, current_height + vdist,
            '{:>6.2%}'.format(table.at[table.index[i], '# peptides in bin'] / total),
            fontsize=10, color='dimgrey', ha='center')

    plt.ylim(0, max_height * 1.2)
    plt.tight_layout()
    plt.savefig(os.path.join(save_directory, 'summary.png'))  # dpi=500
    plt.savefig(os.path.join(save_directory, 'summary.svg'))
    plt.close()



import pandas as pd
import numpy as np
import os
import sys

from collections import defaultdict
from scipy.stats import ttest_ind

import logging
import warnings
from pyteomics import parser
from . import utils, localization, html, io, stats, recommendations

logger = logging.getLogger(__name__)


def get_peptide_statistics(peptide_list):
    '''
    Calculates presence of amino acid in peptide sequences.

    Parameters
    ----------
    peptide_list : Iterable
        An iterable of peptides, that are already fully cleaved.

    Returns
    -------
    out : dict with amino acids as a key and its persentage of peptides with it as a value.
    '''
    sum_aa = 0
    pep_set = set(peptide_list)
    d = defaultdict(int)
    for seq in pep_set:
        for let in set(seq):
            d[let] += 1
        sum_aa += 1
    for i in d:
        d[i] = int(100 * d[i] / sum_aa)
    return d


def get_aa_distribution(peptide_list, rule):
    '''
    Calculates amino acid statistics for peptide list.
    In silico cleaves peptides to get fully cleaved set of peptides.

    Parameters
    ----------
    peptide_list : Iterable
        An iterable of peptides.
    rule : str or compiled regex.
        Cleavage rule in pyteomics format.

    Returns
    -------
    out : dict with amino acids as a key and its persentage as a value.
    '''
    sum_aa = 0
    pep_set = utils.make_0mc_peptides(peptide_list, rule)
    d = defaultdict(int)
    for seq in pep_set:
        for let in seq:
            d[let] += 1
            sum_aa += 1
    for i in d:
        d[i] /= sum_aa
    return d


def make_table(distributions, number_of_PSMs, mass_shifts, reference_label):
    '''
    Prepares amino acid statistis result table.

    Parameters
    ----------
    distributions : DataFrame
        Amino acids statistics, where indexes are amino acids, columns mass shifts (str)
    number_of_PSMs : Series
        Indexes are mass shifts (in str format) and values are numbers of filtered PSMs
    mass_shifts : dict
        Mass shift in str format (rounded) -> actual mass shift (float)

    Returns
    -------

    A table with mass shifts, psms, amino acid statistics columns.
    '''
    df = pd.DataFrame({'mass shift': [mass_shifts[k] for k in distributions.columns],
                       '# peptides in bin': number_of_PSMs},
                      index=distributions.columns)
    df['# peptides in bin'] = df['# peptides in bin'].astype(np.int64)
    out = pd.concat([df, distributions.T], axis=1)
    out['is reference'] = df.index == reference_label
    return out


def calculate_error_and_p_vals(pep_list, err_ref_df, reference, rule, aas):
    '''
    Calculates p-values and error standard deviation of amino acids statistics
    using bootstraping method.

    Parameters
    ----------
    pep_list : Iterable
        An iterable of peptides.
    err_ref_df : Series
        Indexes are amino acids and values are stds of a `reference` mass shift.
    reference : Series
        Indexes are amino acids and values are amino acids statistics of a reference mass shift.
    rule : str or compiled regex.
        Cleavage rule in pyteomics format.
    aas: Iterable
        An Iterable of amino acids to be considered.

    Returns
    -------

    Series of p-values, std of amino acid statistics for considered `pep_list`.
    '''
    d = pd.DataFrame(index=aas)
    for i in range(50):
        d[i] = pd.Series(get_aa_distribution(
            np.random.choice(np.array(pep_list),
            size=(len(pep_list) // 2), replace=False), rule)) / reference
    p_val = pd.Series()
    for i in aas:
        p_val[i] = ttest_ind(err_ref_df.loc[i, :], d.loc[i, :])[1]
    return p_val, d.std(axis=1)


def get_zero_mass_shift(mass_shift_data_dict, params_dict):
    """
    Shift of non-modified peak. Finds zero mass shift.

    Parameters
    ----------
    mass_shift_data_dict : dict
        dict of mass shifts.
    params_dict: dict

    Returns
    -------
    Mass shift label, Mass shift in float format.
    """
    values = [v[0] for v in mass_shift_data_dict.values()]
    keys = list(mass_shift_data_dict.keys())
    data = [v[1] for v in mass_shift_data_dict.values()]
    lref = np.argmin(np.abs(values))
    maxbin = max(df.shape[0] for df in data)
    logger.debug('Closest to zero: %s, with %d peptides. Top mass shift has %d peptides.',
                 keys[lref], data[lref].shape[0], maxbin)
    if abs(values[lref]) > params_dict['zero bin tolerance'] or data[lref].shape[0] / maxbin < params_dict['zero min intensity']:
        logger.warning('Too few unmodified peptides. Mass shift with most identifications will be the reference.')
        identifications = [df.shape[0] for df in data]
        lref = np.argmax(identifications)
    return keys[lref], values[lref]


def check_difference(shift1, shift2, tolerance=0.05):
    """
    Checks two mass shifts means to be closer than the sum of their std.

    Parameters
    ----------
    shift1 : List
        list that describes mass shift. On the first position have to be mean of mass shift,
        on  second position have to be std.
    shift2 : List
        list that describes mass shift. On the first position have to be mean of mass shift,
        on  second position have to be std.
    tolerance : float
        Matching tolerance in Da.

    Returns
    -------
    out : bool
    """
    mean_diff = (shift1[1] - shift2[1]) ** 2
    sigma_diff = (shift1[2] + shift2[2]) ** 2
    res = mean_diff > sigma_diff
    if abs(shift1[1] - shift2[1]) < tolerance:
        res = False
    return res


def filter_mass_shifts(results, tolerance=0.05):
    """
    Merges close mass shifts. If difference between means of two mass shifts less
    than sum of sigmas, they are merged.

    Parameters
    ----------
    results : numpy array
        Output of utils.fit_peaks function (poptperr matrix). An array of Gauss fitted mass shift
        parameters and their tolerances. [[A, mean, sigma, A_error, mean_error, sigma_error],...]
    tolerance : float
        Matching tolerance in Da.
    Returns
    -------
    Updated poptperr matrix.
    """
    logger.info('Discarding bad peaks...')
    temp = []
    out = []
    if not results.size:
        return []
    if results.size == 1:
        return [results[0]]

    temp = [results[0]]
    for mass_shift in results[1:]:
        if check_difference(temp[-1], mass_shift, tolerance=tolerance):
            if len(temp) > 1:
                logger.info('Joined mass shifts %s', ['{:0.4f}'.format(x[1]) for x in temp])
            out.append(max(temp, key=lambda x: x[0]))
            temp = [mass_shift]
        else:
            temp.append(mass_shift)
    out.append(max(temp, key=lambda x: x[0]))

    logger.info('Peaks for subsequent analysis: %s', len(out))
    return out


def calculate_statistics(mass_shifts_dict, reference_label, params_dict, args):
    """
    Calculates amino acid statistics, relative amino acids presence in peptides
    for all mass shifts.

    Paramenters
    -----------
    mass_shifts_dict : dict
        A dict with mass shifts (in str format) as key and values is a DF with filtered PSMs.
    zero_mass_shift : float
        Reference mass shift.
    params_dict : dict
        Dict with paramenters for parsing csv file.
        `labels`, `rule`, `peptides_column` and other params

    Returns
    -------

    DF with amino acid statistics, Series with number of PSMs and dict of data
    for mass shift figures.

    """
    logger.info('Calculating distributions...')
    labels = params_dict['labels']
    rule = params_dict['rule']
    expasy_rule = parser.expasy_rules.get(rule, rule)
    save_directory = args.dir
    peptides = params_dict['peptides_column']
    reference_bin = mass_shifts_dict[reference_label][1]

    number_of_PSMs = dict()  # pd.Series(index=list(mass_shifts_labels.keys()), dtype=int)
    reference = pd.Series(get_aa_distribution(reference_bin[peptides], expasy_rule))
    reference.fillna(0, inplace=True)

    # bootstraping for errors and p values calculation in reference (zero) mass shift
    err_reference_df = pd.DataFrame(index=labels)
    for i in range(50):
        err_reference_df[i] = pd.Series(get_aa_distribution(
            np.random.choice(np.array(reference_bin[peptides]), size=(len(reference_bin) // 2), replace=False),
            expasy_rule)) / reference

    logger.info('Mass shifts:')
    distributions = pd.DataFrame(index=labels)
    p_values = pd.DataFrame(index=labels)

    figure_args = {}

    for ms_label, (ms, ms_df) in mass_shifts_dict.items():
        aa_statistics = pd.Series(get_aa_distribution(ms_df[peptides], expasy_rule))
        peptide_stat = pd.Series(get_peptide_statistics(ms_df[peptides]), index=labels)
        number_of_PSMs[ms_label] = len(ms_df)
        aa_statistics.fillna(0, inplace=True)
        distributions[ms_label] = aa_statistics / reference
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            p_vals, errors = calculate_error_and_p_vals(ms_df[peptides], err_reference_df, reference, expasy_rule, labels)
        # errors.fillna(0, inplace=True)

        p_values[ms_label] = p_vals
        distributions.fillna(0, inplace=True)

        figure_args[ms_label] = (len(ms_df), [distributions[ms_label], errors], peptide_stat.fillna(0))
        logger.info('%s Da', ms_label)

    pout = p_values.T
    pout.fillna(0).to_csv(os.path.join(save_directory, 'p_values.csv'), index=False)
    return distributions, pd.Series(number_of_PSMs), figure_args


def systematic_mass_shift_correction(mass_shifts_dict, mass_correction):
    """

    Parameters
    ----------
    mass_shifts_dict : dict
        A dict with in the format: `mass_shift_label`: `(mass_shift_value, filtered_peptide_dataframe)`.
    mass_correction: float
        Mass of reference (zero) mass shift, that should be moved to 0.0

    Returns
    -------
    out : dict
        Updated `mass_shifts_dict`
    """
    out = {}
    for k, v in mass_shifts_dict.items():
        corr_mass = v[0] - mass_correction
        out[utils.mass_format(corr_mass)] = (corr_mass, v[1])
    return out


def AA_stat(params_dict, args, step=None):
    """
    Calculates all statistics, saves tables and pictures.
    """
    save_directory = args.dir

    logger.debug('Fixed modifications: %s', params_dict['fix_mod'])
    logger.debug('Variable modifications: %s', params_dict['var_mod'])
    logger.info('Using fixed modifications: %s.', utils.format_mod_dict(utils.masses_to_mods(params_dict['fix_mod'])))
    logger.info('Variable modifications in search results: %s.', utils.format_mod_list(params_dict['var_mod']))
    data = io.read_input(args, params_dict)
    if data is None:
        sys.exit(1)

    hist, popt_pvar = stats.fit_peaks(data, args, params_dict)
    # logger.debug('popt_pvar: %s', popt_pvar)
    final_mass_shifts = filter_mass_shifts(popt_pvar, tolerance=params_dict['shift_error'] * params_dict['bin_width'])
    # logger.debug('final_mass_shifts: %s', final_mass_shifts)
    mass_shift_data_dict = utils.group_specific_filtering(data, final_mass_shifts, params_dict)
    # logger.debug('mass_shift_data_dict: %s', mass_shift_data_dict)
    if not mass_shift_data_dict:
        html.render_html_report(None, mass_shift_data_dict, None, params_dict, {}, {}, {}, [], save_directory, [], step=step)
        return None, None, None, mass_shift_data_dict, {}

    reference_label, reference_mass_shift = get_zero_mass_shift(mass_shift_data_dict, params_dict)
    if abs(reference_mass_shift) < params_dict['zero bin tolerance']:
        logger.info('Systematic mass shift equals %s', reference_label)
        if params_dict['calibration'] != 'off':
            mass_shift_data_dict = systematic_mass_shift_correction(mass_shift_data_dict, reference_mass_shift)
            reference_mass_shift = 0.0
            reference_label = utils.mass_format(0.0)
        else:
            logger.info('Leaving systematic shift in place (calibration disabled).')
    else:
        logger.info('Reference mass shift is %s', reference_label)
    ms_labels = {k: v[0] for k, v in mass_shift_data_dict.items()}
    logger.debug('Final shift labels: %s', ms_labels.keys())

    distributions, number_of_PSMs, figure_data = calculate_statistics(mass_shift_data_dict, reference_label, params_dict, args)

    table = make_table(distributions, number_of_PSMs, ms_labels, reference_label)

    stats.summarizing_hist(table, save_directory)
    logger.info('Summary histogram saved.')
    # table.index = table['mass shift'].apply(utils.mass_format)
    table[['is isotope', 'isotope index']] = utils.find_isotopes(
        table['mass shift'], table['# peptides in bin'], tolerance=params_dict['prec_acc'])
    table.at[reference_label, 'is isotope'] = False
    table.at[reference_label, 'isotope index'] = None
    logger.debug('Isotopes:\n%s', table.loc[table['is isotope']])
    u = utils.UNIMOD.mods
    unimod_df = pd.DataFrame(u)
    table['unimod candidates'], table['unimod accessions'] = zip(*table['mass shift'].apply(
        lambda x: utils.get_candidates_from_unimod(x, params_dict['prec_acc'], unimod_df)))

    table['sum of mass shifts'] = utils.find_sums(table.loc[~table['is isotope'], 'mass shift'],
            tolerance=params_dict['shift_error'] * params_dict['bin_width'])
    logger.debug('Sums of mass shifts:\n%s', table.loc[table['sum of mass shifts'].notna()])
    table.to_csv(os.path.join(save_directory, 'aa_statistics_table.csv'), index=False)

    spectra_dict = io.read_spectra(args)

    if spectra_dict:
        if args.mgf:
            params_dict['mzml_files'] = False
        else:
            params_dict['mzml_files'] = True
        logger.info('Starting Localization using MS/MS spectra...')
        ms_labels = pd.Series(ms_labels)
        locmod_df = table[['mass shift', '# peptides in bin', 'is isotope', 'isotope index', 'sum of mass shifts',
            'unimod candidates', 'unimod accessions']].copy()

        locmod_df['aa_stat candidates'] = localization.get_candidates_from_aastat(
            table, labels=params_dict['labels'], threshold=params_dict['candidate threshold'])

        if params_dict['use_all_loc']:
            logger.info('Localizaing all mass shifts on all amino acids. This may take some time.')
            locmod_df['all candidates'] = [set(parser.std_amino_acids)] * locmod_df.shape[0]
        else:
            locmod_df['all candidates'] = locmod_df.apply(
                lambda x: set(x['unimod candidates']) | set(x['aa_stat candidates']), axis=1)
            if params_dict['force_term_loc']:
                logger.debug('Adding terminal localizations for all mass shifts.')
                locmod_df['all candidates'] = locmod_df['all candidates'].apply(lambda x: x | {'N-term', 'C-term'})
        for i in locmod_df.loc[locmod_df['is isotope']].index:
            locmod_df.at[i, 'all candidates'] = locmod_df.at[i, 'all candidates'].union(
                locmod_df.at[locmod_df.at[i, 'isotope index'], 'all candidates'])
        for i in locmod_df.index:
            ac = locmod_df.at[i, 'all candidates']
            for term in ('N', 'C'):
                if 'Protein {}-term'.format(term) in ac and '{}-term'.format(term) in ac:
                    ac.remove('Protein {}-term'.format(term))
                    logger.debug('Removing protein %s-term localization for %s as redundant.', term, i)
        if reference_mass_shift == 0.0:
            locmod_df.at[reference_label, 'all candidates'] = set()
        locmod_df['candidates for loc'] = localization.get_full_set_of_candidates(locmod_df)

        logger.info('Reference mass shift %s', reference_label)
        localization_dict = {}

        for ms_label, (ms, df) in mass_shift_data_dict.items():
            localization_dict.update(localization.localization(
                df, ms, ms_label, locmod_df.at[ms_label, 'candidates for loc'],
                params_dict, spectra_dict, {k: v[0] for k, v in mass_shift_data_dict.items()}))

        locmod_df['localization'] = pd.Series(localization_dict).apply(dict)
        locmod_df.to_csv(os.path.join(save_directory, 'localization_statistics.csv'), index=False)

        if not locmod_df.at[reference_label, 'all candidates']:
            logger.debug('Explicitly writing out peptide table for reference mass shift.')
            df = mass_shift_data_dict[reference_label][1]
            io.save_df(reference_label, df, save_directory, params_dict)
        for reader in spectra_dict.values():
            reader.close()
    else:
        locmod_df = None
        io.save_peptides(mass_shift_data_dict, save_directory, params_dict)
        logger.info('No spectrum files. MS/MS localization is not performed.')
    logger.info('Plotting mass shift figures...')
    for ms_label, data in figure_data.items():
        if locmod_df is not None:
            localizations = locmod_df.at[ms_label, 'localization']
            sumof = locmod_df.at[ms_label, 'sum of mass shifts']
        else:
            localizations = None
            sumof = None
        stats.plot_figure(ms_label, *data, params_dict, save_directory, localizations, sumof)

    logger.info('AA_stat results saved to %s', os.path.abspath(args.dir))
    utils.internal('Data dict: \n%s', mass_shift_data_dict)
    recommended_fix_mods = recommendations.determine_fixed_mods(figure_data, table, locmod_df, mass_shift_data_dict, params_dict)
    logger.debug('Recommended fixed mods: %s', recommended_fix_mods)
    if recommended_fix_mods:
        logger.info('Recommended fixed modifications: %s.', utils.format_mod_dict_str(recommended_fix_mods))
    else:
        logger.info('Fixed modifications not recommended.')
    recommended_var_mods = recommendations.determine_var_mods(
        figure_data, table, locmod_df, mass_shift_data_dict, params_dict, recommended_fix_mods)
    logger.debug('Recommended variable mods: %s', recommended_var_mods)
    if recommended_var_mods:
        logger.info('Recommended variable modifications: %s.', utils.format_mod_list(recommended_var_mods))
    else:
        logger.info('Variable modifications not recommended.')
    combinations = utils.get_varmod_combinations(recommended_var_mods, ms_labels, params_dict['prec_acc'])
    logger.debug('Found combinations in recommended variable mods: %s', combinations)
    opposite = utils.get_opposite_mods(
        params_dict['fix_mod'], recommended_fix_mods, recommended_var_mods, ms_labels, params_dict['prec_acc'])
    logger.debug('Opposite modifications: %s', utils.format_mod_list([recommended_var_mods[i] for i in opposite]))
    html.render_html_report(table, mass_shift_data_dict, locmod_df, params_dict,
        recommended_fix_mods, recommended_var_mods, combinations, opposite, save_directory, ms_labels, step=step)
    return figure_data, table, locmod_df, mass_shift_data_dict, recommended_fix_mods, recommended_var_mods



import argparse
import logging
import os
from . import AA_stat, utils, io


def main():
    pars = argparse.ArgumentParser(epilog='Instead of file lists, you can pass directory names. '
        'This will process all files in the directory.')
    pars.add_argument('--params', help='CFG file with parameters. If there is no file, AA_stat uses default one. '
        'An example can be found at https://github.com/SimpleNumber/aa_stat', required=False)

    pars.add_argument('--dir', help='Directory to store the results. Default value is current directory.', default='.')
    pars.add_argument('-v', '--verbosity', type=int, choices=range(4), default=1, help='Output verbosity.')

    input_spectra = pars.add_mutually_exclusive_group()
    input_spectra.add_argument('--mgf', nargs='+', help='MGF files to localize modifications.')
    input_spectra.add_argument('--mzml', nargs='+', help='mzML files to localize modifications.')

    input_file = pars.add_mutually_exclusive_group(required=True)
    input_file.add_argument('--pepxml', nargs='+', help='List of input files in pepXML format.')
    input_file.add_argument('--csv', nargs='+', help='List of input files in CSV format.')

    args = pars.parse_args()
    levels = [logging.WARNING, logging.INFO, logging.DEBUG, utils.INTERNAL]
    logging.basicConfig(format='{levelname:>8}: {asctime} {message}',
                        datefmt='[%H:%M:%S]', level=levels[args.verbosity], style='{')

    # Performance optimizations as per https://docs.python.org/3/howto/logging.html#optimization
    logging._srcfile = None
    logging.logThreads = 0
    logging.logProcesses = 0
    logger = logging.getLogger(__name__)
    logging.getLogger('matplotlib').setLevel(logging.WARNING)

    logger.info('Starting...')

    io.resolve_filenames(args)
    params_dict = io.get_params_dict(args)
    logger.debug(params_dict)

    os.makedirs(args.dir, exist_ok=True)
    AA_stat.AA_stat(params_dict, args)
    logger.info('Done.')
